(*
	knowledge.ml
	23/01/2003
	Giacomo Baldi

	Implementation of a knowledge base to represent the intruder knowledge.
*)

open Message
open Utils


type  kb = int*(msg list)


let void =
  function () -> (-1,[])


let applyall =
  function (kbid,k) ->
    function asg -> (kbid, map k (asg)) 

let expand =
  function (kbid,a) ->
    function n -> 
      function asg -> 
	let (ll, nk) =  applyall (kbid,a) asg in
	(n,nk)

let get =
  function (kbid,k) ->
    function f -> (filter k f)

let rec scan =
  function (kbid,k) ->
    function p -> exists k p

let knows =
  function k ->
      function m ->
	scan k (function mm ->  mm = m)

let knowsd =
  function k ->
    function asg ->
      function m ->
	scan k (function mm ->  (asg mm) = m)

let add =
  function ((kbid,k), n) -> (kbid, n::k)

let addlist =
  function ((kbid,k), n) -> (kbid, n@k)

let rec getlist =
  function (kbid, k) -> k

let rec getAllCrypts =
  function k ->
    get k (iscrypt)


let rec listfy =
  function (k,ln) ->
    let rec llistfy =
      function n ->
	(
	 match n with
	 | Couple(m1, m2) -> (llistfy m1)@(llistfy m2)
	 | Crypt(mm, K(s,i)) when knows k (K(s,i))  -> llistfy mm
	 | Crypt(mm, PNp(s,p)) when knows k  (PNm(s,p))  -> if (knows k (PNp(s,p))) then llistfy mm else n::(llistfy mm)
	 | Crypt(mm, PNm(s,p)) when knows k  (PNp(s,p))  -> if (knows k (PNm(s,p))) then llistfy mm else n::(llistfy mm)
	 | Crypt(mm, SVar(_,_,_)) -> llistfy mm
	 | Crypt(mm, Publ(SVar(_,_,_))) -> llistfy mm
	 | Crypt(mm, Priv(SVar(_,_,_))) -> llistfy mm
	 | Crypt(mm,kk) when not(knows k (n)) -> [n]
	 | K(_) when not (knows k n) -> [n]
	 | PNp(s,p) when not (knows k n) -> [n]
	 | PNm(s,p) when not (knows k n) -> [n]
	 | NO(s,p) when not (knows k n) -> [n]
	 | PN(s,p) when not (knows k n) -> [n]
	 | SVar(_,_,_) when not (knows k n) -> [n]
	 | Var(_) when not (knows k n) -> [n]
	 | _ -> []
	) in
    match ln with
    | m1::ms -> (llistfy(m1))@(listfy(k,ms))
    | [] -> []

let rec learn =
  function (k,n) ->
    if n = Bottom then k else (
    let gg = remclone (listfy(k,n::(getAllCrypts k))) (function m1 -> function m2 -> m1<>m2) in
    let kk = filter gg (function m -> (match m with | Crypt(_,_) -> false | _ -> true)&&(not (knows k m))) in
    let cc = filter gg (function m -> (match m with | Crypt(_,_) -> true  | _ -> false)&&(not (knows k m))) in
    if kk = [] then addlist(k,cc)
    else learn (addlist(k,kk),n)
   )

let rec derives =
  function k ->
    function asg ->
      function m ->
	(match m with
	| Couple(m1,m2) ->  (derives k (asg) m1) && (derives k (asg) m2)
	| Crypt(m1,K(s,i)) -> (derives k (asg) m1) && (derives k (asg) (K(s,i)))
	| Crypt(m1,PNp(a,b)) -> (derives k (asg) m1) && (derives k (asg) (PNm(a,b)))
	| Crypt(m1,PNm(a,b)) -> (derives k (asg) m1) && (derives k (asg) (PNp(a,b)))
	| SVar(v,i,_) -> true
	| o -> knowsd k asg o
	)||(knowsd k asg m)



let getIntruderVariables =
  function k ->
    get k (function m -> match m with | Var(_) -> true | _ -> false)


let rec getAllNames =
  function k ->
    get k (isname)

let rec getRNames =
  function k ->
      get k (function m -> (isname m) && (knows k (toPNp m)) && (knows k (toPNm m)))

let isRName =
  function k ->
    function n ->
      let rn = getRNames k in
      exists rn (function m -> m=n)



let rec getNames =
  function k ->
      diff (getAllNames k) (getRNames k)


let rec getKeys =
  function k ->
    get k (function m -> isskey m || ismkey m || ispkey m )

let rec getSKeys =
  function k ->
    get k isskey


let rec getAllPKeys =
  function k ->
    get k ispkey

let rec getPMKeys =
  function k ->
    get k (function m ->  (ispkey m && (knows k (compl(m)))))
      

let rec getPKeys =
  function k ->
    diff (getAllPKeys k) (getPMKeys k)


let isPMKey =
  function k ->
    function n ->
      let rn = getPMKeys k in
      exists rn (function m -> toName(m)=toName(n))

let rec getAllMKeys =
  function k ->
      get k ismkey

let rec getMPKeys =
  function k ->
      map (getPMKeys k) (function m -> compl(m))

let rec getMKeys =
  function k ->
      diff (getAllMKeys k) (getMPKeys k)


let rec getCrypts =
  function k ->
    function kk ->
      if (knows k  (compl(kk))) && (knows k  kk) then []
      else (
	get k (function m ->
	  (match m with
	  | Crypt(_,ss) when (ss=kk)-> true
	  | _ -> false
	  )
	      )    
       )

let rec getSCrypts =
  function k ->
      let keys = getSKeys k in
      let res = map keys (function kk -> getCrypts k  kk) in
      debox res

let rec getPCrypts =
  function k ->
      let keys = getPKeys k in
      let res = map keys (function kp -> getCrypts k kp) in
      debox res

let rec getMCrypts =
  function k ->
      let keys = getMKeys k in
      let res = map keys (function km -> getCrypts k km) in
      debox res


let id_of =
  function (kbid,k) -> kbid


let isEmpty =
  function (kbid, k) -> k=[]

let isVoid =
  function (kbid, k) -> kbid= -1


let rec hasKeys =
  function  k ->
    scan k (function m -> (isskey m) || (ispkey m) || (ismkey m))

let rec hasSKeys =
  function  k ->
    scan k isskey

let rec hasPKeys =
  function  k ->
    scan k ispkey

let rec hasMKeys =
  function  k ->
    scan k ismkey

let rec hasNames =
  function  k ->
    scan k isname

let rec hasPMKeys =
  function  k ->
    scan k (
    function m ->
      if (ispkey m || ismkey m) then (
	scan k (function mm -> mm = compl(m))
       ) else false
   )

let rec hasRNames =
  function  k ->
    scan k (
    function m -> if isname m then (
      ((get k (function m1 -> m1 = toPNp m))<>[])&&((get k (function m1 -> m1 = toPNm m))<>[])
     ) else false
   )



let rec prntHtml =
  function (kbid, k) ->
    let r = ref(0) in
    let rec prntKb =
      function ml ->
	match ml with
	| m::xs -> Message.prntMsgHtml m; print_string(", "); r:=!r+1; (if (!r mod 5=0) then print_string("<br>") else ()); prntKb xs
	| [] -> (print_string(".<br>"))
    in
    (prntKb k)


let rec prnt =
  function (kbid, k) ->
    let rec prntKb =
      function ml ->
	match ml with
	| m::xs -> Message.prntMsg m; print_string("\n"); prntKb xs
	| [] -> (print_string(".\n"))
    in
    (print_string("Kid: "^string_of_int(kbid)^"\n"); prntKb k)
      
let rec tostring =
  function (kbid,k) ->
    let rec prntKb =
      function ml ->
	match ml with
	| m::[] -> (Message.tostring m)^(prntKb [])
	| m::xs -> (Message.tostring m)^(", ")^(prntKb xs)
	| [] -> (".\\n\\n")
    in
    ("("^string_of_int(kbid)^")"^prntKb k)
      













