(*
	Principal.ml
	30/01/2003
	Giacomo Baldi
	
	Implementation of a principal.

*)


open Process
open Message
open Utils
	
(* 
	A principal is made of a name, an index, a list of open variables and a process expression.
*)

type principal = string*int*(variable list)*process


(* create a principal*)
let create =
  function (name, index, vlist, expr) -> 
    let rec updateVars =
      function ls ->
	match ls with
	| x::xs -> setVariable((name,index),x)::(updateVars xs)
	| [] -> []
    in
    let rec updateMsg =
      function m ->
	match m with
	| Couple(m1,m2) -> Couple(updateMsg m1, updateMsg m2)
	| Crypt(m1,m2) -> Crypt(updateMsg m1, updateMsg m2)
	| (Var(v)) -> (Var(setVariable((name,index),v)))
	| (BVar(v)) -> (BVar(setVariable((name,index),v)))
	| Publ(Var(v)) -> Publ(Var(setVariable((name,index),v)))
	| Priv(Var(v)) -> Priv(Var(setVariable((name,index),v)))
	| (NO(s,p)) -> (NO(s,index))
	| (PNp(s,p)) -> (PNp(s,index))
	| (PNm(s,p)) -> (PNm(s,index))
	| (PN(s,p)) -> (PN(s,index))
	| (K(s,i)) -> (K(s,index))
	| o -> o
    in
    let rec updateExpr =
      function e ->
	match e with
	| Par(e1,e2) -> Par(updateExpr e1, updateExpr e2)
	| Sum(e1,e2) -> Sum(updateExpr e1, updateExpr e2)
	| In(m,e1) -> In(updateMsg m, updateExpr e1)
	| Out(m,e1) -> Out(updateMsg m, updateExpr e1)
	| Nil -> Nil
    in (name, index, updateVars vlist, updateExpr expr)

let check =
  function (pn,pi,pv,pp) ->
    Process.validate pp

let substitute =
  function asg ->
    function (pn,pi,pv,pp) ->
      let rec subvar =
	function vv ->
	  match vv with
	  | v::vs -> if (asg (Var(v)))<>Var(v) then subvar vs else v::(subvar vs)
	  | [] -> []
      in
      (pn,pi, subvar pv, Process.substitute asg pp)

let typeOf =
  function (pn,pi,pv,pp) ->
    function v ->
      let rec findtype =
	function pr ->
	  match pr with
	  | In(m,e) -> 
	      let tt = (typeOfVar m v) in 
	      if tt  = Gen then findtype e 
	      else tt
	  | Out(m,e) ->
	      let tt = (typeOfVar m v) in 
	      if tt  = Gen then findtype e 
	      else tt
	  | Par(e1,e2) -> 
	      let tt1 = findtype e1 in
	      let tt2 = findtype e2 in
	      mintype(tt1,tt2)
	  | Sum(e1,e2) -> 
	      let tt1 = findtype e1 in
	      let tt2 = findtype e2 in
	      mintype(tt1,tt2)
	  | Nil -> Gen
      in
      if exists pv (function vv -> v=vv) then findtype pp
      else Gen
	  

let isConsumed =
  function (_,_,v,e) -> (e=Nil && v=[])

let getName =
  function (n,i,_,_) -> (n,i)

let getVars = 
  function (_,_,vs,_) -> map vs (function v -> Var(v))

let len = 
  function (_,_,_,e) -> Process.len e
      
let prnt =
  function p ->
    function asg ->
      let (name, index, vlist, expr) = substitute asg p in (
      print_string(name^"_"^string_of_int(index)^": ");
      print_string("(");
      Message.prntVarlist vlist;
      print_string(") [");
      Process.prnt (expr);
      print_string("]")
	
     )

let prntHtml =
  function p ->
    function asg ->
      let (name, index, vlist, expr) = substitute asg p in (
      print_string(name^"<sub>"^string_of_int(index)^"</sub>: ");
      print_string("(");
      Message.prntVarlist vlist;
      print_string(") [");
      Process.prntHtml (expr);
      print_string("]")
	
     )

let tostring =
  function p ->
    function asg ->
      let (name, index, vlist, expr) = substitute asg p in
      (""^name^"_"^string_of_int(index)^": (")^((Message.tostringVarList vlist)^") ["^(Process.tostring (expr))^"]")

