(*
	utils.mli
	01/05/2003
	Giacomo Baldi

	Definition of utility functions
*)
(**
	This module contains utility functions used to perform operations
	over lists and over higher-order function values.
*)

(**
	raised when an error occurs.
*)
exception UtilException


(**
    Len takes a list l.
	@return the number of elements in a list
*)
val len: 'a list -> int

(**
	Select takes a list l and an integer n.
	@return the nth element in list
	@raise UtilException if list is empty
*)
val select: 'a list -> int -> 'a

(**
	Map takes a list l and a function f.
	@return a list where each element is the result of the application of f to the elements of l.
*)
val map: 'a list -> ('a -> 'b) -> 'b list

(**
	Iter takes a list l and a function f.
	f is applied to each element in l.
*)
val iter: 'a list -> ('a -> unit) -> unit 

(**
	Filter takes a list l and a predicate p.
	@return a list that contains the elements of l verifying p
*)
val filter: 'a list -> ('a -> bool) -> 'a list

(**
	Debox takes a list of lists l.
	@return a list that is the concatenation of the lists in l.
*)
val debox: 'a list list -> 'a list

(**
	Revers takes a list l.
	@return a list that is the reverse of l.
*)
val reverse: 'a list -> 'a list

(**
	Mul takes a list l1, a list l2 and a function f.
	It generates the cartesian product (cp) between l1 and l2 and then
	@return a mapping of cp with function f
*)
val mul: 'a list -> 'b list -> ( 'a -> 'b -> 'c ) -> 'c list

(**
	Remclone takes a list l, and a binary predicate p.
	@return a filtering of l where each element is compared with each other using p.
*)
val remclone: 'a list -> ('a -> 'a ->  bool) -> 'a list

(**
	Diff takes a list l1, and a list l2.
	@return a list containing the elements of l1 not in l2
*)
val diff: 'a list -> 'a list -> 'a list


(**
	grow takes an element a and a integer n.
	@return a list of n repetitions of a.
*)
val grow: 'a -> int -> 'a list

(**
   couplify takes two lists l1 and l2.
   @return a list of couples of element of l1 and l2
   @raise UtilException when lists aren't of the same length
*)
val couplify: 'a list -> 'b list -> ('a * 'b) list


(**
	Aleph takes a list l.
	@return a list of all the subsets of the elements in l
*)
val aleph: 'a list -> 'a list list

(**
	Fold takes a list l, function f and a base element e.
	@return a value of the same type of e, obtained by recursively iterate the application
	of f to the elements of l. f ( f ( f xn e) xn-1) ... x1)
	@raise UtilException if l is emtpy or  a singleton
*)
val fold: 'a list -> ('a -> 'b -> 'b) -> 'b -> 'b

(**
	Fold2 takes a list l, function f.
	@return a value obtained by recursively iterate the application
	of f to the elements of l. f ( f ( f xn-1 xn) xn-2) ... x1)
*)
val fold2: 'a list -> ('a -> 'a -> 'a) -> 'a


(**
	Head takes a list l.
	@return  the first element of l.
	@raise UtilException if l is empty.
*)
val head: 'a list -> 'a

(**
	Tail takes a list l.
	@return l without the first element.
*)
val tail: 'a list -> 'a list


(**
	sort takes a list l and a predicate p to compare elements of l.
	@return a sorted list
*)
val sort: 'a list -> ('a -> 'a -> bool) -> 'a list

(**
	Exists takes a list l and a predicate p.
	@return true if exists an element of l that verifies p, false otherwise.
*)
val exists: 'a list -> ('a -> bool) -> bool

(**
	Forall takes a list l and a predicate p.
	@return true if every element of l verifies p, false otherwise.
*)
val forall: 'a list -> ('a -> bool) -> bool

(**
	Used internally
*)
val rebuild: 'a list -> 'b list list -> ( 'a -> 'b -> 'c ) -> 'c list

(**
    Used Internally
*)
val tensor: 'a list -> ('a -> 'a) -> 'a list list

(**
   Partition take a list l and
   @return a list of partitions of elements in l
*)
val partition: 'a list -> 'a list list list

(**
	Print_list takes a list l and a printing function pr.
	It iterates l printing each element with pr.
*)
val print_list: 'a list -> ('a -> unit) -> unit

(*functions over functions*)

(**
	Compose takes a function f and a function g.
	@return the composition of f and g, which is a function obtained by first applying f and then g.
*)
val compose: ('a -> 'b) -> ('b -> 'c) -> 'a -> 'c

(**
	Invert takes a binary function f.
	@return a function g that takes parameters in reverse order with respect to f.
*)
val invert: ('a -> 'b -> 'c) -> ('b -> 'a -> 'c)

(**
	half takes a binary function f.
	@return a function g that takes half the parameters with respect to f.
*)
val half: ('a -> 'a -> 'b) -> ('a -> 'b)
