(*
	Logic.ml
	14/03/2003
	Giacomo Baldi

	Implementation of a protocol logic representation

*)



open Message
open Node
open Utils
open Assignment


type index = string

(*
	A formula in protocol logic (PL) is a:
	- Universally quantified formula with an index ranging over a specific principal
	- Existentially quantified formula with an index ranging over a specific principal
	- Standard boolean operations between formulas
	- Equality (Inequality) among variable and datum
	- Derivability (Non Derivability) of a message
*)

type pls =
  | Forall of index*string*pls
  | Exists of index*string*pls
  | Not of pls
  | And of pls*pls
  | Or of pls*pls
  | Equal of msg*msg
  | Diff of msg*msg
  | Derive of msg
  | NDerive of msg
  | True
  | False


type npls =
    pls list list

(*
   Given a formula p we expand it removing quantifiers and reducing
   p to a disjunction of conjucts
 *)


let rec not_norm =
  function w ->
    match w with
    | Not(And(a,b)) -> Or(not_norm(Not(a)), not_norm(Not(b)))
    | Not(Or(a,b)) -> And(not_norm(Not(a)), not_norm(Not(b)))
    | Not(Not(a)) -> not_norm a
    | Not(Equal(v,m)) -> Diff(v,m)
    | Not(Diff(v,m)) -> Equal(v,m)
    | Not(Derive(m)) -> NDerive(m)
    | Not(NDerive(m)) -> Derive(m)
    | Not(Forall(i,s,a)) -> Exists(i,s,not_norm(Not(a)))
    | Not(Exists(i,s,a)) -> Forall(i,s,not_norm(Not(a)))
    | Not(True) -> False
    | Not(False) -> True
    | And(a,b) -> And(not_norm a, not_norm b)
    | Or(a,b) -> Or(not_norm a, not_norm b)
    | Forall(i,s,a) -> Forall(i,s, not_norm a)
    | Exists(i,s,a) -> Exists(i,s, not_norm a)
    | _ -> w

let  and_or_norm =
  function p ->
    let i = ref(0) in
    let rec and_or = 
      function w ->
	match w with
	| And(Or(b,c),a)
	| And(a,Or(b,c)) -> 
	    let z = and_or (And(a,b)) in
	    let v = and_or(And(a,c)) in
	    (i:= !i+1; Or(z,v))
	| And(a,b) -> 
	    let v1 = !i in
	    let c = and_or a in
	    let v2 = !i in 
	    let d = and_or b in
	    if (v1 = v2 && v2 = !i) then And(c,d) else and_or (And(c,d))
	| Or(a,b) -> Or(and_or a, and_or b)
	| _ -> w
    in
    and_or p
    


let rec expand_quantifiers =
  function w ->
    function ct ->
      let rec expand =
	function is ->
	  function i ->
	    function name ->
	      function l ->
		let rec ist =
		  function mm ->
		    match mm with
		    | Var(n,(p,ii)) when p=is -> Var(n,(name,i))
		    | Publ(Var(n,(p,ii))) when p=is -> Publ(Var(n,(name,i)))
		    | Priv(Var(n,(p,ii))) when p=is -> Priv(Var(n,(name,i)))
		    | NO(n,ii) when n.[0]='_' ->
			let pll = (String.rindex n '_' ) in
			let nl = String.length n in
			let iii = (String.sub n 1 (pll-1)) in
			if iii=is then NO(String.sub n (pll+1) ((nl)-pll-1), i)
			else mm
		    | K(n,ii) when n.[0]='_' ->
			let pll = (String.rindex n '_' ) in
			let nl = String.length n in
			let iii = (String.sub n 1 (pll-1)) in
			if iii=is then K(String.sub n (pll+1) ((nl)-pll-1), i)
			else mm
		    | PN(n,ii) when n=(name^is) -> PN(name,i)
		    | PNm(n,ii) when n=(name^is) -> PNm(name,i)
		    | PNp(n,ii) when n=(name^is) -> PNp(name,i)
		    | Couple(m1,m2) -> Couple(ist m1, ist m2)
		    | Crypt(m1,kk) -> Crypt(ist m1, ist kk)
		    | _ -> mm
		in
		match l with
		| And(a,b) -> And(expand is i name a, expand is i name b)
		| Or(a,b) -> Or(expand is i name a, expand is i name b)
		| Forall(ii,s,a) -> Forall(ii, s, expand is i name a)
		| Exists(ii,s,a) -> Exists(ii, s, expand is i name a)
		| Equal(v,m) -> Equal(ist v, ist m)
		| Diff(v,m) -> Diff(ist v, ist m)
		| Derive(m) -> Derive(ist m)
		| NDerive(m) -> NDerive(ist m)
		| _ -> l
      in
      match w with
      | Forall(i,s,a) ->
	  let b = expand_quantifiers a ct in
	  let ii = Context.getIndexesOf ct s in
	  let tt = map ii (function index -> expand i index s b) in
	  if ((len tt) < 2)  then (if tt=[] then True else head tt)
	  else fold2 tt (function x -> function y -> And(x,y))
      | Exists(i,s,a) ->
	  let b = expand_quantifiers a ct in
	  let ii = Context.getIndexesOf ct s in
	  let tt = map ii (function index -> expand i index s b) in
	  if ((len tt) < 2)  then (if tt=[] then False else head tt)
	  else fold2 tt (function x -> function y -> Or(x,y))
      | And(a,b) -> And(expand_quantifiers a ct, expand_quantifiers b ct)
      | Or(a,b) -> Or(expand_quantifiers a ct, expand_quantifiers b ct)
      | _ -> w

let rec listfy =
  function pl ->
    let rec listfyAnd =
      function pa ->
	match pa with
	| And(a,b) -> (listfyAnd a)@(listfyAnd b)
	| _ -> [pa]
    in
    let rec listfyOr =
      function po ->
	match po with
	| Or(a,b) -> (listfyOr a)@(listfyOr b)
	| And(_,_) -> [listfyAnd po]
	| _ -> [[po]]
    in
    listfyOr pl



let rec prnt =
  function p ->
    match p with
    | Forall(i,n,pl) -> (print_string("{FA "^i^":"^n^"| "); prnt pl; print_string("}"))
    | Exists(i,n,pl) -> (print_string("{EX "^i^":"^n^"| "); prnt pl; print_string("}"))
    | Not(pl) -> (print_string("[not "); prnt pl; print_string("]"))
    | And(pl1,pl2) -> (print_string("{"); prnt pl1; print_string(" and "); prnt pl2; print_string("}"))
    | Or(pl1,pl2) -> (print_string("["); prnt pl1; print_string(" or "); prnt pl2; print_string("]"))
    | Equal(v,m) -> (print_string("("); prntMsg v; print_string(" = "); prntMsg m; print_string(")"))
    | Derive(m) -> (print_string("( K :> "); prntMsg m; print_string(")"))
    | Diff(v,m) -> (print_string("("); prntMsg v; print_string(" <> "); prntMsg m; print_string(")"))
    | NDerive(m) -> (print_string("( K !:> "); prntMsg m; print_string(")"))
    | True -> (print_string(" T "))
    | False -> (print_string(" F "))

let rec prntHtml =
  function p ->
    match p with
    | Forall(i,n,pl) -> (print_string("Forall "^n^"<sub>"^i^"</sub>: "); prntHtml pl)
    | Exists(i,n,pl) -> (print_string("Exists "^n^"<sub>"^i^"</sub>: "); prntHtml pl)
    | Not(pl) -> (print_string("not "); prntHtml pl)
    | And(pl1,pl2) -> (prntHtml pl1; print_string(" and "); prntHtml pl2)
    | Or(pl1,pl2) -> (prntHtml pl1; print_string(" or "); prntHtml pl2)
    | Equal(v,m) -> (prntMsgHtml v; print_string(" = "); prntMsgHtml m)
    | Derive(m) -> (print_string("K :&gt "); prntMsgHtml m)
    | Diff(v,m) -> (prntMsgHtml v; print_string(" &lt &gt "); prntMsgHtml m)
    | NDerive(m) -> (print_string("K !:&gt "); prntMsgHtml m)
    | True -> (print_string(" T "))
    | False -> (print_string(" F "))


let prntnplHtml =
  function pln->
    print_list pln (
    function ll ->
      print_list ll (
      function pll -> prntHtml pll; print_string("<br>")
     ); print_string("<br>")
   )

let prntnpl =
  function pln->
    print_list pln (
    function ll ->
      print_list ll (
      function pll -> prnt pll; print_string("\n")
     ); print_string("\n")
   )

let rec normalize =
  function p ->
    function node ->
      let checkequality =
	function a1 ->
	  function a2 ->
	    let rec compare =
	      function l1 ->
		function l2 ->
		  match (l1) with
		  | ll::ls ->
		      if exists l2 (function y -> y=ll) then compare ls l2
		      else false
		  | [] -> true
	    in
	    if (len a1) = (len a2) then (
	      (compare a1 a2) && (compare a2 a1)
	     )
	    else false
      in 
      let (ct,asg,k) = Node.getstatus node in
      let ttt = and_or_norm(expand_quantifiers (not_norm p) ct) in
      let res = remclone (listfy ttt) (function x -> function y -> not (checkequality x y)) in
      (*print_string("****\n");Context.prnt ct asg; prntnpl res;print_string("----\n");*)res

let rec concretize =
  function pl ->
    function asg ->
      let subst =
	function y ->
	  map y (function z ->
	    match z with
	    | Equal(v,m) -> Equal(asg v, asg m)
	    | Diff(v,m) -> Diff(asg v, asg m)
	    | Derive(m) -> Derive(asg m)
	    | NDerive(m) -> NDerive(asg m)
	    | _ -> z
		)
      in
      map pl (function x -> subst x)


let getatoms =
function n1 ->
function n2 ->
let ttt = couplify n1 n2 in
let res = map ttt (
		   function (a,b) -> 
let trr= couplify a b in 
let atoms = filter trr (function (a1, a2) -> a2=False && (a1<>False)) in
map atoms (fst)
) in remclone (debox res) (<>)
