(*
	Node.ml
	30/01/2003
	Giacomo Baldi

	Implementation of a node.
*)

open Status
open Principal
open Message
open Assignment
open Process
open Action

type redex = principal*action*principal

(*
	A node is the fundamental structure to handle search in state space. A node is made of
	a status, a reference to an int (nd), a reference to a node (father), a reference to a label,
	an int (id).
		- status: is the status (see status.ml) owned by the node
		- nd:
		- father: is a reference to the node from which the transition that create the current node, has been made.
		- label: is a label to remember what kind of action brings from father to current node
		- id: is used for debugging purpose and is the unique identifier of a node.
	A void node is a NulNode.
*)
type noderecord = {
    mutable status: status;
    mutable numprinc: int;
    mutable numjoin: int;
    mutable father: node;
    mutable label: action;
    mutable redextable: redex list;
    mutable jointable: action list;
    mutable closed: bool;	
    mutable inorout: bool;
    mutable joining: bool;
    mutable nsons: int;
    mutable id: int;
    mutable level: int
	
}
and node = NulNode | Node of noderecord

let kid = ref(0)

(* create a new node incrementing the global kid *)
let create =
  function st -> kid:=!kid+1;
    let nn = {
      status = st;
      numprinc = 0;
      numjoin = 0;
      father = NulNode;
      label = NoMoreAct;
      redextable = [];
      jointable = [];
      closed = false;
      inorout = true;
      joining = true;
      nsons = 0;
      id = !kid;
      level = 0;
    } in Node(nn)
      

(* Node is an abstract type so we need some function to access data *)

let setstatus =
  function Node(nn),stl -> nn.status <- stl

let setlabel =
  function Node(nn),lbl -> nn.label <- lbl

let settable =
  function Node(nn),rll -> nn.redextable <- rll

let setjoins =
  function Node(nn),jjr -> nn.jointable <- jjr

let addson =
  function Node(nn) ->
    nn.nsons <- (nn.nsons+1)



let setnd =
  function Node(nn),npp -> nn.numprinc <- npp

let setnj =
  function Node(nn),njj -> nn.numjoin <- njj

let close =
  function Node(nn) -> nn.closed <- true

let setIn =
  function Node(nn) -> nn.inorout <- false

let setJn =
  function Node(nn) -> nn.joining <- false

let getlevel =
  function Node(nn) -> (nn.level)
  
let setfather =
  function Node(nn),fth -> (
    nn.father <- fth;
    nn.level <- ((getlevel fth)+1);
    addson fth
   )



let getlabel =
  function Node(nn) -> (nn.label)

let getfather =
  function Node(nn) -> (nn.father)

let getnd =
  function Node(nn) -> (nn.numprinc)

let getnj =
  function Node(nn) -> (nn.numjoin)

let getstatus =
  function Node(nn) -> (nn.status)

let gettable =
  function Node(nn) -> (nn.redextable)

let getjoins =
  function Node(nn) -> (nn.jointable)

let getSons =
  function Node(nn) -> (nn.nsons)


let getRedex =
  function Node(nn) ->
    match (nn.redextable) with
    | x::xs -> (nn.redextable <- xs; x)
    | [] when (nn.closed) -> (("",0,[],Nil),Closed,("",0,[],Nil))
    | [] -> (("",0,[],Nil),NoMoreAct,("",0,[],Nil))

let getJoin =
  function Node(nn) ->
    match (nn.jointable) with
    | x::xs -> (nn.jointable <- xs; x)
    | [] when nn.closed -> (Closed)
    | [] -> (NoMoreAct)


let getid =
  function Node(nn) -> nn.id
    | _ -> -1

let isNul =
  function NulNode -> true
    | Node (_) -> false

let isOut =
  function NulNode -> true
    | Node (nn) -> nn.inorout

let isJn =
  function NulNode -> true
    | Node (nn) -> nn.joining

let isLeaf =
  function NulNode -> false
    | Node (nn) -> (nn.nsons = 0)


let nulNode =
  function () -> NulNode


let isTerminal =
  function Node(nn) -> Status.isTerminal (nn.status)

let isPartiallyTerminated =
  function Node(nn) -> Status.isPartiallyTerminated (nn.status)




let rec prntT =
  function ll ->
    match ll with
    | [] -> print_string("\n");
    | (p1,Input(pp,m,s),p2)::ls -> (
	print_string("Input <");
	prntMsg (s m);
	print_string(">\n");
	prntT ls
       )
    | (p1,Output(pp,m,s),p2)::ls -> (
	print_string("Output <");
	prntMsg (s m);
	print_string(">\n");
	prntT ls
       )
let rec prntJ =
  function ll ->
    match ll with
    | [] -> print_string("\n");
    | (Join(p,s))::ls -> (
	print_string("Join <");
	print_string(fst(p)^"_"^string_of_int(snd(p)));
	print_string(">\n");
	prntJ ls
       )

let prnt =
  function Node(nn) ->
    (
     print_string("\n---------------------------------\nAction: ");
     Action.prnt (nn.label);
     print_string("\n   ||\n   ||\n   \\/\n");
     print_string("\n-----\nNode: id="^string_of_int(nn.id)^" lev="^string_of_int(nn.level)^"\n-----\nSon of:"^string_of_int(getid(nn.father))^"\n");
     Status.prnt (nn.status)
(*		print_string("table:\n");
   prntT (nn.redextable);
   print_string("table:\n");
   prntJ (nn.jointable)
 *)
    )
    | NulNode -> print_string("\nNode: NULNODE\n")
