(* Parser *)
open Tokenizer
open Message
open Principal
open Process

(*
	This parser is used to extract a principal from a text file given the following sintax:
	
	<Principal>::= <Name>: (<VarList>) [<Process>];
	
	<VarList>::= lambda || <Variable>,<Vars> || <Variable>
	<Vars>::= <Variable> || <Variable>,<Vars>

	<Process>::= lambda ||
		|| in(<Message>)
		|| out(<Message>)
		|| (<Process>)
		|| <Process>.<Process>
		|| <Process> + <Process>
		|| <Process> | <Process>

	<Message>::= (<Message>, <Message>)
		|| {<Message>}<Message>
		|| <Key>
		|| <PName>
		|| <Pname>+
		|| <Pname>-
		|| <Nonce>
		|| <Variable>
		|| ?<Variable>
		|| <Variable>+
		|| <Variable>-

	<Name>::= an uppercase string representing the name of the principal
	<Key>::= a string beginning with 'k'
	<PName>::= an uppercase string representin principal names or keys if followed by (+,-)
	<Nonce>::= a string beginning with 'n'
	<Variable>::= a lowercase string not beginning with 'k' or 'n'
	
	
	- Summation has an higher precedence with respect to Parallelism
	(i.e. in(x)+out(y) | out(z)+in(w) means (in(x)+out(y)) | (out(z)+in(w)) )

	- Only variables and decorated pnames can occur as a key in a crypt
	
	- Comments are allowed everywhere. A comment can be a '//' and means that
	everything on that line is ignored. A multiline comment begina with '/*'
	and end with '*/'. Everything in between is ignored.
	

	Refer to den_sacco for an example of input file.
*)


(*
	Internally used.
	- Pexp represent an expression already formed
	- Dot represent the '.' operator
	- Plp stands for an open par '('
	- Ino represents an input action
	- Outo represents an output action
	- Sumo represents a summation
	- Paro represents a parallel
	- End is end of parsing
*)

type pars =
	| Pexp of process
	| Dot
	| Plp
	| Ino of msg
	| Outo of msg
	| Sumo
	| Paro
	| End

(*
	ParseError is raised when a general error occurs
	CantReduce is raised when trying to reduce expression on the stack
*)

exception ParseError
exception CantReduce

(*
	error codes.
*)

let errOK							= 0;;
let errCOMMENT_NOT_OPENED			= 1;;
let errCOMMENT_NOT_CLOSED			= 2;;
let errUNKNOWN_TOKEN				= 3;;
let errMISSING_OPENING_ROUND_PAR	= 4;;
let errMISSING_CLOSING_ROUND_PAR	= 5;;
let errMISSING_OPENING_SQUARE_PAR	= 6;;
let errMISSING_CLOSING_SQUARE_PAR	= 7;;
let errMISSING_OPENING_CURLY_PAR	= 8;;
let errMISSING_CLOSING_CURLY_PAR	= 9;;
let errMISSING_COMMA				= 10;;
let errMISSING_SEMICOLON			= 11;;
let errMISSING_COLON				= 12;;
let errEXPECTED_IDE					= 13;;
let errUNEXPECTED_TOKEN				= 14;;
let errWRONG_KEY					= 15;;
let errWRONG_VAR					= 16;;
let errMALFORMED_EXPR				= 17;;
let errUPPERCASE_PRINCIPAL			= 18;;
let errBINDING_VARIABLE_AS_A_KEY	= 19;;

(*
	error string table.
*)

let errorString = Array.create 20 "";;
errorString.(errOK) <- "Ok.";;
errorString.(errCOMMENT_NOT_OPENED) <- "Comment not opened";;
errorString.(errCOMMENT_NOT_CLOSED) <- "Comment not closed";;
errorString.(errUNKNOWN_TOKEN) <- "Unknown token";;
errorString.(errMISSING_OPENING_ROUND_PAR) <- "Missing (";;
errorString.(errMISSING_CLOSING_ROUND_PAR) <- "Missing )";;
errorString.(errMISSING_OPENING_SQUARE_PAR) <- "Missing [";;
errorString.(errMISSING_CLOSING_SQUARE_PAR) <- "Missing ]";;
errorString.(errMISSING_OPENING_CURLY_PAR) <- "Missing {";;
errorString.(errMISSING_CLOSING_CURLY_PAR) <- "Missing }";;
errorString.(errMISSING_COMMA) <- "Missing ,";;
errorString.(errMISSING_SEMICOLON) <- "Missing ;";;
errorString.(errMISSING_COLON) <- "Missing :";;
errorString.(errEXPECTED_IDE) <- "Identifier expected";;
errorString.(errUNEXPECTED_TOKEN) <- "Unexpected token";;
errorString.(errWRONG_KEY) <- "Wrong key";;
errorString.(errWRONG_VAR) <- "Wrong variable";;
errorString.(errMALFORMED_EXPR) <- "Malformed expression";;
errorString.(errUPPERCASE_PRINCIPAL) <- "Principal name should be uppercase";;
errorString.(errBINDING_VARIABLE_AS_A_KEY) <- "A binding variable cannot occur as a key";;


(*
	returns a string representing the line and the column of the error position
*)
let whereError =
	function () -> " ("^(string_of_int(fst(where())))^","^(string_of_int(snd(where())))^")"


let errorCode = ref(0)
let errorData = ref ("")
let errorMsg =
	function () -> "ER"^string_of_int(!errorCode)^": "^errorString.(!errorCode)^" "^(!errorData)^" "^whereError()^"\n"


(*
	used to signal an error while parsing
*)
let putError =
	function (code, s) -> errorData:= s; errorCode:=code

let getErrorMsg = function () -> errorMsg()
let getErrorCode = function () -> !errorCode



(*
	check if a string is uppercase
*)
let isUpper =
	function s ->
		let isUp =
			function ch ->
				match ch with
					| 'A'..'Z' -> true
					| _ ->  false
		in
		let rec isU =
			function (s, i, m) ->
				if (i<m) then
					if isUp(s.[i]) then isU(s, i+1, m)
					else false
				else true
		in isU(s,0, String.length s)


(*
	reads from tokenizer until current comment is closed
*)
let rec ignoreComment =
	function () ->
		let a = getToken() in
			match a with
				| LSym("*") ->
					let b = getToken() in
						if b=LSym("/") then ()
						else (pushback(b); ignoreComment())
				| LStop -> (raise ParseError)
				| _ -> ignoreComment()


(*
	extracts current token taking care of comments
*)
let rec readToken =
	function () ->
		let a = getToken() in
			match a with
				| LSym("/") ->
					let b = getToken() in
						if b=LSym("/") then (flushline(); readToken())
						else
							if b=LSym("*") then ((let ss = whereError() in try (ignoreComment()) with ParseError -> putError(errCOMMENT_NOT_CLOSED,ss)); readToken())
							else (pushback(b); a)
				| LSym("*") ->
					let b = getToken() in
						if b=LSym("/") then (putError(errCOMMENT_NOT_OPENED,""); raise ParseError)
						else (pushback(b); a)
				| LUnknown(a) -> (putError(errUNKNOWN_TOKEN,""); raise ParseError)
				| _ -> a


(*
	This is a bunch of functions to parse messages.
	After parsing a 'in' or a 'out' we start parsing message calling getMsg.
	We search for a '(' then we parse the message  with getMessage and search for a ')'.

	While parsing with getMessage we scan the input for :
	- '(' which means we expect a couple
	- '{' which means we expect a crypt
	- '?' which means we expect a binding var
	- identifier, which means we expect something as a variable or a key.

	To parse a couple we first parse the first message, then we search for a ','
	then we parse the second message expecting a final matching ')'.

	To parse a crypt we first parse the message, then we search for '}' then we parse
	the key.

	getVal is used to recognize variables, names, nonces and keys.
	If we have an uppercase identifier it is a principal name or an asymmetric key
	if it is followed by '+' or '-'.
	If we have an identifier starting with 'k' it is a key.
	If we have an identifier starting with 'n' it is a nonce.
	Every other identifier is a variable which can be decorated
*)

let rec getMsg =
	function () ->
		let p1 = readToken() in
			if p1<>LSym("(") then (putError(errMISSING_OPENING_ROUND_PAR,""); raise ParseError)
			else
				let ms = getMessage() in
					let p2 = readToken() in
						if p2<>LSym(")") then ( putError(errMISSING_CLOSING_ROUND_PAR,""); raise ParseError)
						else ms
and getMessage =
	function() ->
	let a = readToken() in
		match a with
			| LSym("(") -> getCouple()
			| LSym("{") -> getCrypt()
			| LSym("?") -> (
				let y=getVal() in
					match y with
						| (Var(s,(a,b))) -> (BVar(s,(a,b)))
						| Publ(v)
						| Priv(v) -> (putError(errBINDING_VARIABLE_AS_A_KEY,""); raise ParseError)
						| _ -> (putError(errUNKNOWN_TOKEN,""); raise ParseError)
				)
			| LIde(v) -> pushback(a); getVal()
			| _ -> (putError(errUNKNOWN_TOKEN,""); raise ParseError)
and getCouple =
	function () ->
		let m1 = getMessage() in
			let v1 = readToken() in
				if v1<>LSym(",") then (putError(errMISSING_COMMA,""); raise ParseError)
				else
					let m2 = getMessage() in
						let v2 = readToken() in
							if v2<>LSym(")") then (putError(errMISSING_CLOSING_ROUND_PAR,""); raise ParseError)
							else Couple(m1,m2)
and getCrypt =
	function () ->
		let m1 = getMessage() in
			let v1 = readToken() in
				if v1<>LSym("}") then (putError(errMISSING_CLOSING_CURLY_PAR,""); raise ParseError)
				else
					let k1 = getVal() in Crypt(m1,k1)
and getVal =
	function () ->
		let a = readToken() in
			match a with
				| LIde(s) when isUpper(s) -> (
					let k = readToken() in
						match k with
							| LSym("+") -> (PNp(s,0))
							| LSym("-") -> (PNm(s,0))
							| LSym(c) -> pushback(k); (PN(s,0))
							| _ ->(putError(errWRONG_KEY,""); raise ParseError)
						)
				| LIde(s) -> (
					if s.[0] = 'n' then (NO(s,0))
					else
						if s.[0] = 'k' then (K(s,0))
						else (
							let k = readToken() in
								match k with
									| LSym("+") -> Publ(Var(s,("",0)))
									| LSym("-") -> Priv(Var(s,("",0)))
									| LSym(c) -> pushback(k); (Var(s,("",0)))
									| _ ->(putError(errWRONG_VAR,""); raise ParseError)
								)
					)
				| _ -> (putError(errEXPECTED_IDE,""); raise ParseError);;



(*
	while putting things on the parsing stack with parseExp, we use reducer
	trying to pack the current stack.
	- if we saw a lonely action we pack it in a Pexp
	- if we saw a Pexp Dot Action we pack it all in a corrispondig Pexp
	- if we saw a summation between two Pexp we pack it because it hase maximum precedence
	- if we saw a parallel we pack it only if the priority of this reduction is different from Summation
*)
let reducer =
	function (stack, pr) ->
		match stack with
			| Ino(m)::st -> Pexp(In(m,Nil))::st
			| Outo(m)::st -> Pexp(Out(m,Nil))::st
			| Pexp(e)::Dot::Outo(m1)::st -> Pexp(Out(m1,e))::st
			| Pexp(e)::Dot::Ino(m1)::st -> Pexp(In(m1,e))::st
			| Pexp(e1)::Sumo::Pexp(e2)::st -> Pexp(Sum(e2,e1))::st
			| Pexp(e1)::Paro::Pexp(e2)::st when pr<>Sumo -> Pexp(Par(e2,e1))::st
			| _ -> raise CantReduce;;

(*
	This is the heart of process parsing.
	Given a token and a stack (initially empty) we analyze
	the token and
	- if we have an action we first parse the relative message and put the action on the stack
	- if we have a '.' we put Dot on the stack
	- if we have a '+' it means we have to handle a summation. So we try to reduce the stack (knowing
	what is the current precedence) and we iterate until we can't reduce. At this point we have
	only to put Sumo on the stack knowing that it is reduced. Precedence is respected because
	every parallel operation int he stack isn't reduced until we pack all summation in between.
	- if we have a '|' we can reduce everything without problems
	- if we have a '(' we put a Plp in the stack and we will use it when we get a ')'
	- if we have a ']' we are at the end of process

	In parsing [in(x) + out(y) | in(z)] we have:

	- tok=in		stack=[]
	- tok='+'		stack=Ino(x) reduce to stack=Pexp(In(x,Nil))
	- tok=out		stack=Sumo::Pexp(...)
	- tok='|'		stack=out(y)::Sumo::Ino(x) and we reduce to stack=Pexp(Sum(in(x),out(y))) respecting precedence
	- tok=in		stack=Ino(z)::Paro::Pexp(...)
	- tok=End		stack=End::Ino(z)::Paro::Pexp(...)

	next we'll call reduce_all for the last reduction
*)
let rec parseExp =
	function (tok, stack) ->
		match (tok, stack) with
			| LIde("in"), _ -> let ms = getMsg() in Ino(ms)::stack
			| LIde("out"), _ -> let ms = getMsg() in Outo(ms)::stack
			| LSym("."), _ -> Dot::stack
			| LSym("+"), _ -> (try parseExp (tok, reducer(stack, Sumo)) with CantReduce -> Sumo::stack)
			| LSym("|"), _ -> (try parseExp (tok, reducer(stack, Paro)) with CantReduce -> Paro::stack)
			| LSym("("), _ -> Plp::stack
			| LSym(")"), Pexp(e)::Plp::st -> Pexp(e)::st
			| LSym(")"), _ -> parseExp (tok, reducer(stack, Dot))
			| LSym("]"), _ -> End::stack
			| _ -> putError(errUNKNOWN_TOKEN,""); raise ParseError;;


let rec reduce_all =
	function stack ->
		match stack with
			| [Pexp(e)] -> e
			| [] -> (putError(errMISSING_SEMICOLON,"");Nil)
			| _ -> reduce_all ( reducer (stack, Dot));;


(*
	Calls parseExp for each token until end of input or end of parsing. At this point we returns
	the current stack without the final End.
*)
let rec parser =
	function stack ->
		let a = readToken() in
			(match a with
				| LStop -> []
				| _ ->
					let ss = parseExp (a, stack) in
					let c = List.hd ss in
						if c=End then stack
						else parser ss
			);;



(*
	Parsing of <name>. We search for an uppercase string and a following colon.
*)
let parseName =
	function () ->
		let a = readToken() in
			match a with
				| LIde(n) when isUpper(n) -> (
					let s = readToken() in
						if s<>LSym(":") then (putError(errMISSING_COLON,""); raise ParseError)
						else n
					)
				| _ -> (putError(errUPPERCASE_PRINCIPAL,""); raise ParseError);;

(*
	Parsing of <VarList>
*)
let rec parseVariables =
	function () ->
		let p1 = readToken() in
			if p1<>LSym("(") then (putError(errMISSING_OPENING_ROUND_PAR,""); raise ParseError)
			else
				let vs = parseVars([]) in
					let p2 = readToken() in
						if p2<>LSym(")") then (putError(errMISSING_CLOSING_ROUND_PAR,""); raise ParseError)
						else vs
and parseVars =
	function ll ->
		let a = readToken() in
			match (a,ll) with
				| (LSym(")"), _) -> pushback(a); ll
				| (LIde(s), _ ) -> (
					let a2 = readToken() in
						if (a2<> LSym(",") && a2<>LSym(")") ) then (putError(errUNEXPECTED_TOKEN,""); raise ParseError)
						else (pushback(a2); parseVars(ll@[(s,("",0))]))
					)
				| (LSym(","), l::ls) -> (
					let a2 = readToken() in
						match a2 with
							| LIde(j) -> (pushback(a2); parseVars(ll))
							| _ -> (putError(errUNEXPECTED_TOKEN,""); raise ParseError)
					)
				| _ -> (putError(errUNEXPECTED_TOKEN,""); raise ParseError);;


(*
	We parse <name>, <VarList> and <Process> creating the corrisponding principal
*)
let parsePrincipal =
	function () ->
		let (fl1, name) = (try ((true,parseName())) with ParseError -> (false,""))
		in
			if not fl1 then Principal.create("",-1, [], Nil)
			else
				let (fl2,vars) = (try ((true,parseVariables())) with ParseError -> (false,[]))
				in
					if not fl2 then Principal.create("",-1, [], Nil)
					else
						let p1 = readToken() in
							if p1<>LSym("[") then (putError(errMISSING_OPENING_SQUARE_PAR,""); Principal.create("",-1, [], Nil))
							else
								let (fl3,ex) = (
									try ((true, reduce_all (parser []))) with
										| CantReduce -> (putError(errMALFORMED_EXPR,""); (false,Nil))
										| ParseError -> (false, Nil)
									)
								in
									if fl3 then
										let p2 = readToken() in
											if p2<>LSym(";") then (putError(errMISSING_SEMICOLON,""); Principal.create("",-1, [], Nil))
											else (putError(errOK,""); Principal.create(name, 0, vars, ex))
									else Principal.create("",-1, [], Nil)


let more_parsing =
	function () ->
		let a = readToken() in (pushback(a); a <> LStop)

(*
	Starts to read from fname.
	Close fname.
	Theese are used in setup and shutdown phase.
*)
let init_parser =
	function fname -> init_tokenizer(fname);;
let done_parser =
	function () -> done_tokenizer();;

