(*
	process.ml
	23/01/2003
	Giacomo Baldi

	Operations on process representation
*)



open Message
open Assignment


(*
	The Behavioural expression of a Principal is a Process. A Process is:
		- an action of Input(Mv) or Output(Mv) followed by a process expression
		- a summation of processes
		- a parallel of processes
		- no action

	** Definition of Mv is in Message.ml **
*)

exception SubstitutionException

type process =
	| In of msg*process
	| Out of msg*process
	| Sum of process*process
	| Par of process*process
	| Nil


(*
	Substitution over a process expressions is simplified by underlying hypothesis:
		- Principal are closed: no free variables
		- a binding variable can appear only once in a principal expression


	Given the assignment (v,a), for each variable v in process e, a substitution
	is a new process where every occurence of v is replaced by its value a.

	Given the assignment (v,a), for each variable v in message mm, a substitution
	is a new ground message where every occurence of v is replaced by its value a.
	A decorated variable (Publ or Priv) is replaced with the right key.

*)

let rec substitute =
	function asg ->
		function e ->
			match e with
				| Par(e1, e2) -> Par(substitute (asg) e1, substitute (asg) e2)
				| Sum(e1, e2) -> Sum(substitute (asg) e1, substitute (asg) e2)
				| In(m,e1) -> In(asg m, substitute (asg) e1)
				| Out(m,e1) -> Out(asg m, substitute (asg) e1)
				| Nil -> Nil


(* to do *)

let rec validate =
	function e ->
		match e with
			| Par(e1, e2)
			| Sum(e1, e2) -> (validate e1)&&(validate e2)
			| In(m,e1)
			| Out(m,e1) -> (validate e1)&&(checkmsg m)
			| Nil -> true

let rec len =
	function e ->
		match e with
			| Par(e1, e2) -> (len e1) + (len e2)
			| Sum(e1, e2) -> 
				let l1 = len e1 in
				let l2 = len e2 in
					if (l1>l2) then l1 else l2
			| In(m,e1)
			| Out(m,e1) -> 1+(len e1)
			| Nil -> 0

let rec tostring =
  function s ->
    match s with
    | In(m,Nil) -> ("In(")^(Message.tostring m)^")"
    | Out(m,Nil) -> ("Out(")^(Message.tostring m)^")"
    | In(m,e) -> ("In(")^(Message.tostring m)^")."^(tostring e)
    | Out(m,e) -> ("Out(")^(Message.tostring m)^")."^(tostring e)
    | Sum(e1,e2) -> "("^(tostring e1)^" + "^(tostring e2)^")"
    | Par(e1,e2) -> ("(")^(tostring e1)^" | "^(tostring e2)^")"
    | Nil -> ""

let rec prnt =
  function s ->
    match s with
    | In(m,Nil) -> print_string(("In(")^(Message.tostring m)^")")
    | Out(m,Nil) ->  print_string(("Out(")^(Message.tostring m)^")")
    | In(m,e) ->  print_string(("In(")^(Message.tostring m)^")."^(tostring e))
    | Out(m,e) ->  print_string(("Out(")^(Message.tostring m)^")."^(tostring e))
    | Sum(e1,e2) ->  print_string("("^(tostring e1)^" + "^(tostring e2)^")")
    | Par(e1,e2) ->  print_string(("(")^(tostring e1)^" | "^(tostring e2)^")")
    | Nil -> ()

let rec prntHtml =
  function s ->
    match s with
    | In(m,Nil) -> print_string(("<i>in</i>(")); prntMsgHtml m; print_string (")")
    | Out(m,Nil) ->  print_string(("<i>out</i>(")); prntMsgHtml m; print_string (")")
    | In(m,e) ->  print_string(("<i>in</i>(")); prntMsgHtml m; print_string (")."); prntHtml e
    | Out(m,e) ->  print_string(("<i>out</i>(")); prntMsgHtml m; print_string (")."); prntHtml e
    | Sum(e1,e2) ->  print_string("("); prntHtml e1; print_string(" + "); prntHtml e2; print_string(")")
    | Par(e1,e2) ->  print_string("("); prntHtml e1; print_string(" | "); prntHtml e2; print_string(")")
    | Nil -> ()

