(*
	knowledge.mli
	23/01/2003
	Giacomo Baldi

	Definition of knowledge base
*)
(**
	This module implements the intruder knowledge. A knowledge is a set of messages with
	some operations associated, like membership or derivability.
*)

open Message
open Assignment


(**
	kb is an abstract type. It is internally implemented as a list of messages.
*)
type kb
	
(**
	knows takes a knowledge k, an assignment a and a message m.
	@return true if a(m) is in k, false otherwise.
*)
val knows: kb ->  msg -> bool

(**
	learn takes a knowledge k and a message m.
	@return a knowledge equal to k plus m.
*)
val learn: kb*msg -> kb

(**
	void takes no parameters
	@return an empty knowledge.
*)
val void: unit -> kb

(**
	expand takes a knowledge k and an integer id.
	@return a knowledge with identification id and with all the messages in k.
*)
val expand: kb -> int -> assignment -> kb

(**
	addlist takes a knowledge k and a message list ml.
	@return a knowledge with all the messages in k plus ml.
*)
val addlist: kb*(msg list) -> kb

(**
	derives takes a knowledge k, an assignment a and a message m.
	@return true if m can be derived in k under assignment a.
*)
val derives: kb -> assignment -> msg -> bool

(**
	id_of takes a knowledge k.
	@return the identification of k.
*)
val id_of: kb -> int

(**
   applyall takes a knowledge k and an assignment a.
   @return a new knowledge with all messages processed by a.
*)
val applyall: kb -> assignment -> kb

(**
	isEmpty takes a knowledge k.
	@return true if k is empty, false otherwise.
*)
val isEmpty: kb -> bool

(**
	isVoid takes a knowledge k.
	@return true if k is not a valid knowledge, false otherwise.
*)
val isVoid: kb -> bool

(**
	The following functions are used to retrieve values from a knowledge. They all
	takes a knowledge k anf  return a message list containing all
	the messages in k that match a retrieving criterion.
*)


(**

   @return all open variables of the intruder
*)
val getIntruderVariables: kb -> msg list

(**
	@return all cryptogram in k with a specified key.
*)
val getCrypts: kb ->  msg -> msg list

(**
	@return all cryptogram in k with a symmetric key
*)
val getSCrypts: kb ->  msg list

(**
	@return all cryptogram in k with a private key
*)
val getMCrypts: kb ->  msg list

(**
	@return all cryptogram in k with a public key
*)
val getPCrypts: kb ->  msg list

(**
	@return all names in k.
*)
val getAllNames: kb -> msg list

(**
	@return all names n in k such that n+ and n- are in k.
*)
val getRNames: kb ->  msg list

(**
	@return all names n in k such that n+ or n- are not in k.
*)
val getNames: kb ->  msg list

(**
	@return all keys in k.
*)
val getKeys: kb ->  msg list

(**
	@return all symmetric keys in k.
*)
val getSKeys: kb -> msg list

(**
	@return all public keys in k.
*)
val getAllPKeys: kb -> msg list

(**
	@return all private keys in k.
*)
val getAllMKeys: kb -> msg list

(**
	@return all public keys a+ in k such that a- is in k.
*)
val getPMKeys: kb -> msg list

(**
	@return all public keys a+ in k such that a- is not in k.
*)
val getPKeys: kb ->  msg list

(**
	@return all private keys a- in k such that a+ is in k.
*)
val getMPKeys: kb -> msg list

(**
	@return all private keys a- in k such that a+ is not in k.
*)
val getMKeys: kb -> msg list

(**
	The following functions are used to check if a knowledge contains a certain type of datum.
	They all takes a knowledge k.
*)

(**
	@return true if k contains names.
*)
val hasNames: kb -> bool

(**
	@return true if k contains names and the public and private keys associated.
*)
val hasRNames: kb -> bool

(**
	@return true if k contains keys.
*)
val hasKeys: kb -> bool

(**
	@return true if k contains symmetric keys.
*)
val hasSKeys: kb -> bool

(**
	@return true if k contains public keys.
*)
val hasPKeys: kb -> bool

(**
	@return true if k contains private keys.
*)
val hasMKeys: kb -> bool

(**
	@return true if k contains at least a matching couple of public/private keys.
*)
val hasPMKeys: kb -> bool

(**
	isRName takes a knowledge k and a message m.
	@return true if m is a rname in k, false otherwise.
*)
val isRName: kb -> msg -> bool

(**
	isPMKey takes a knowledge k and a message m.
	@return true if m is a PMKey in k, false otherwise.
*)
val isPMKey: kb -> msg -> bool


(**
	The following functions are used to retrieve and print a representation of the knowledge,
	in a readable format.
*)

(**
	prnt takes a knowledge and prints out the message list.
*)
val prnt: kb -> unit

(**
	prntHtml takes a knowledge and prints out the message list.
*)
val prntHtml: kb -> unit

(**
	tostring takes a knowledge k.
	@return a string representation of messages in k.
*)
val tostring: kb -> string
