(*
	Message.ml
	23/01/2003
	Giacomo Baldi

	Implementation of operations over messages and variables
*)



(*
	A pname is a couple of (name,index) to represent a principal name.
	A variable has a name and refers to a principal.
	A klink is a placeholder for a knowledge used in symbolic variables
	A Message is:
	- a Couple of messages
	- a Crypt of a message with a message
	- a Simmetric Key (K)
	- an Asimmetric Key (PNp, PNm)
	- a Principal Name (PN)
	- a Nonce  (NO)
	- a Variable (Var)
	- a Binding VAriable (BVar)
	- a Symbolic variable (SVar) with a link to a knowledge or a restricted one
	- a decorated Variable (Publ [+], Priv [-])

	A Ground Message (gM) has no variable. If a message has variables its a Message with Variables (Mv)
	If a Message has symbolic variables then its a symbolic message (Ms)
*)

type pname = string*int
type variable = string*pname
type svtype =
  | Gen
  | GenR
  | SKey
  | Name
  | RName
  | PMKey
  | MPKey



type msg =
  | K of string*int
  | PN of string*int
  | PNp of string*int
  | PNm of string*int
  | NO of string*int

  | Var of variable
  | BVar of variable
  | SVar of variable*int*svtype

  | Crypt of msg*msg
  | Couple of msg*msg
  | Publ of msg
  | Priv of msg
  | Bottom
  | Trick of (msg*msg) list




(* modify a variable with a new principal name *)
let setVariable ((name,index),(namevar,(namep, indexp))) = (namevar,(name,index))


(*
   a message, as defined by syntax, could be semantically wrong.
   In fact:
   - all messages of the form Crypt(m1,m2) should be discarded if m2 is not a key or m1 is wrong.
   - only non-binding variables can be decorated
   - binding variables cannot appear as a key in a cryptogram
 *)

let notSVar =
  function m ->
    match m with
    | SVar(_,_,_) -> false
    | _ -> true

let notVar =
  function m ->
    match m with
    | Var(_) -> false
    | _ -> true

let compl =
  function kk ->
    match kk with
    | K(_) -> kk
    | PNp(s,p) -> PNm(s,p)
    | PNm(s,p) -> PNp(s,p)
    | _ -> Bottom

let toName =
  function
    | PNp(p,i) -> PN(p,i)
    | PNm(p,i) -> PN(p,i)
    | _ -> Bottom

let toPNp =
  function
    | PN(p,i) -> PNp(p,i)
    | _ -> Bottom

let toPNm =
  function
    | PN(p,i) -> PNm(p,i)
    | _ -> Bottom

let isKey =
  function k ->
    match k with
    | K(_)
    | PNp(_,_)
    | PNm(_,_)
    | Var(_)
    | SVar(_,_,_)
    | Publ(Var(_))
    | Priv(Var(_))
    | Publ(SVar(_))
    | Priv(SVar(_)) -> true
    | _ -> false

let isGroundKey =
  function k ->
    match k with
    | K(_)
    | PNp(_,_)
    | PNm(_,_) -> true
    | _ -> false

let rec checkmsg =
  function m ->
    match m with
    | Crypt(m1,k) -> (checkmsg m1)&&isKey(k)
    | Couple(m1,m2) -> (checkmsg m1)&&(checkmsg m2)
    | Publ(K(_))
    | Publ(PNp(_))
    | Publ(PNm(_))
    | Priv(K(_))
    | Priv(PNp(_))
    | Priv(PNm(_)) -> false
    | _ -> true


let rec isBinding =
  function (m,v) ->
    match m with
    | Couple(m1,m2) -> isBinding(m1,v)||isBinding(m2,v)
    | Crypt(m1,_) -> isBinding(m1,v)
    | BVar(var) -> Var(var)=v
    | _ -> false

let prntVar =
  function (s,(a,i)) -> print_string(s^"_"^a^"_"^string_of_int(i))

let prntVarHtml =
  function (s,(a,i)) -> print_string(s^"<sub>"^string_of_int(i)^"</sub>")

let tostringVar =
  function (s,(a,i)) -> (s^"_"^a^"_"^string_of_int(i))


let rec prntVarlist =
  function vs -> match vs with
  | x::y::xs -> prntVar x; print_string(", "); prntVarlist (y::xs)
  | x::[] -> prntVar x
  | [] -> ()

let rec prntVarlistHtml =
  function vs -> match vs with
  | x::y::xs -> prntVarHtml x; print_string(", "); prntVarlistHtml (y::xs)
  | x::[] -> prntVarHtml x
  | [] -> ()

let rec tostringVarList =
  function vs -> match vs with
  | x::y::xs -> (tostringVar x)^", "^(tostringVarList (y::xs))
  | x::[] -> tostringVar x
  | [] -> ""


let rec prntMsgHtml =
  function s ->
    match s with
    | (K(m,i)) -> print_string(m^"<sub>"^string_of_int(i)^"</sub>")
    | (PN(s,i)) -> print_string(s^"<sub>"^string_of_int(i)^"</sub>")
    | (PNp(s,i)) -> print_string(s^"<sub>"^string_of_int(i)^"</sub>"^"<sup>+</sup>")
    | (PNm(s,i)) -> print_string(s^"<sub>"^string_of_int(i)^"</sub>"^"<sup>-</sup>")
    | (NO(s,i)) -> print_string (s^"<sub>"^string_of_int(i)^"</sub>")
    | Crypt(m1,m2) -> (print_string "{";prntMsgHtml m1; print_string "}"; print_string "<sub>"; prntMsgHtml m2; print_string "</sub>")
    | Couple(m1,m2) -> (prntMsgHtml m1; print_string ","; prntMsgHtml m2;)
    | (Var(s,(a,i)))
    | (BVar(s,(a,i))) -> prntVarHtml (s,(a,i))
    | (SVar((s,(a,i)),n, Gen)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":*)"))
    | (SVar((s,(a,i)),n, GenR)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":*r)"))
    | (SVar((s,(a,i)),n, SKey)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":sk)"))
    | (SVar((s,(a,i)),n, Name)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":n"))
    | (SVar((s,(a,i)),n, RName)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":nr)"))
    | (SVar((s,(a,i)),n, MPKey)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":mp)"))
    | (SVar((s,(a,i)),n, PMKey)) -> ((prntVarHtml (s,(a,i))); print_string("(K:"^string_of_int(n)^":pm)"))
    | Publ(Var(v)) -> (prntVarHtml v; print_string("<sup>+</sup>"))
    | Priv(Var(v)) -> (prntVarHtml v; print_string("<sup>-</sup>"))
    | Publ(SVar(v,i,t)) -> (prntMsgHtml(SVar(v,i,t)); print_string("<sup>+</sup>"))
    | Priv(SVar(v,i,t)) -> (prntMsgHtml(SVar(v,i,t)); print_string("<sup>-</sup>"))
    | Publ(nn) -> (prntMsgHtml(nn); print_string("<sup>+</sup>"))
    | Priv(nn) -> (prntMsgHtml(nn); print_string("<sup>-</sup>"))
    | Bottom -> print_string(".")
    | Trick(_) -> print_string("Trick")

let rec prntMsg =
  function s ->
    match s with
    | (K(m,i)) -> print_string(m^"_"^string_of_int(i))
    | (PN(s,i)) -> print_string(s^"_"^string_of_int(i))
    | (PNp(s,i)) -> print_string(s^"_"^string_of_int(i)^"+")
    | (PNm(s,i)) -> print_string(s^"_"^string_of_int(i)^"-")
    | (NO(s,i)) -> print_string (s^"_"^string_of_int(i))
    | Crypt(m1,m2) -> (print_string "{";prntMsg m1; print_string "}"; prntMsg m2)
    | Couple(m1,m2) -> (prntMsg m1; print_string ","; prntMsg m2;)
    | (Var(s,(a,i)))
    | (BVar(s,(a,i))) -> prntVar (s,(a,i))
    | (SVar((s,(a,i)),n, Gen)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":gen)"))
    | (SVar((s,(a,i)),n, GenR)) -> ((prntVar (s,(a,i))); print_string("(KR:"^string_of_int(n)^":gen)"))
    | (SVar((s,(a,i)),n, SKey)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":skey)"))
    | (SVar((s,(a,i)),n, Name)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":name)"))
    | (SVar((s,(a,i)),n, RName)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":rname)"))
    | (SVar((s,(a,i)),n, MPKey)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":mpkey)"))
    | (SVar((s,(a,i)),n, PMKey)) -> ((prntVar (s,(a,i))); print_string("(K:"^string_of_int(n)^":pmkey)"))
    | Publ(Var(v)) -> (prntVar v; print_string("+"))
    | Priv(Var(v)) -> (prntVar v; print_string("-"))
    | Publ(SVar(v,i,t)) -> (prntMsg(SVar(v,i,t)); print_string("+"))
    | Priv(SVar(v,i,t)) -> (prntMsg(SVar(v,i,t)); print_string("-"))
    | Publ(nn) -> (prntMsg(nn); print_string("+"))
    | Priv(nn) -> (prntMsg(nn); print_string("-"))
    | Bottom -> print_string(".")
    | Trick(_) -> print_string("Trick")

let rec tostring =
  function s ->
    match s with
    | (K(m,i)) -> (m^"_"^string_of_int(i))
    | (PN(s,i)) ->(s^"_"^string_of_int(i))
    | (PNp(s,i)) -> (s^"_"^string_of_int(i)^"+")
    | (PNm(s,i)) -> (s^"_"^string_of_int(i)^"-")
    | (NO(s,i)) -> (s^"_"^string_of_int(i))
    | Crypt(m1,m2) -> "{"^(tostring m1)^"}"^(tostring m2)
    | Couple(m1,m2) -> ((tostring m1)^","^(tostring m2))
    | (Var(s,(a,i)))
    | (BVar(s,(a,i))) -> tostringVar (s,(a,i))
    | (SVar((s,(a,i)),n,Gen)) -> ((tostringVar (s,(a,i)))^("(KG:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,GenR)) -> ((tostringVar (s,(a,i)))^("(KRG:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,SKey)) -> ((tostringVar (s,(a,i)))^("(KSK:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,Name)) -> ((tostringVar (s,(a,i)))^("(KN:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,RName)) -> ((tostringVar (s,(a,i)))^("(KRN:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,PMKey)) -> ((tostringVar (s,(a,i)))^("(KPM:"^string_of_int(n)^")"))
    | (SVar((s,(a,i)),n,MPKey)) -> ((tostringVar (s,(a,i)))^("(KMP:"^string_of_int(n)^")"))
    | Publ(Var(v)) -> (tostringVar v)^("+")
    | Priv(Var(v)) -> (tostringVar v)^("-")
    | Publ(SVar(v,i,t)) -> (tostring(SVar(v,i,t))^"+")
    | Priv(SVar(v,i,t)) -> (tostring(SVar(v,i,t))^"-")
    | Publ(nn) -> (tostring(nn)^"+")
    | Priv(nn) -> (tostring(nn)^"-")
    | Bottom -> (".")
    | Trick(_) -> ("Trick")

let rec hasntFreeVar =
  function m ->
    let rec checkall =
      function mm ->
	match mm with
	| Couple(m1,m2) -> (checkall m1) && checkall(m2)
	| Crypt(m1, kk) when isGroundKey kk-> (checkall m1)
	| Crypt(m1, Var(v))
	| Crypt(m1, Publ(Var(v)))
	| Crypt(m1, Priv(Var(v))) -> (checkall m1)&&isBinding(m,Var(v))
	| Var(_) -> false
	| Publ(Var(_)) -> false
	| Priv(Var(_)) -> false
	| _ -> true
    in checkall m










let ispkey =
  function m ->
    match m with
    | PNp(_,_) -> true
    | _ -> false

let ismkey =
  function m ->
    match m with
    | PNm(_,_) -> true
    | _ -> false

let isskey =
  function m ->
    match m with
    | K(_) -> true
    | _ -> false

let isname =
  function m ->
    match m with
    | PN(_,_) -> true
    | _ -> false

let iscrypt =
  function m ->
    match m with
    | Crypt(_,_) -> true
    | _ -> false



let mintype =
  function (t1,t2) ->
    let tval =
      function t ->(
	match t with
	| Gen -> 4
	| GenR -> 3
	| Name -> 1
	| SKey -> 1
	| RName -> 0
	| _ -> -1)
    in
    if tval(t1)>tval(t2) then t2 else t1

let comparable =
  function (t1,t2) ->
    ((t1<>Name && t1<>RName)||(t2<>SKey))&&
    ((t2<>Name && t2<>RName)||(t1<>SKey))

let rec typeOfVar =
  function m ->
    function v ->
      match m with
      | Couple(m1,m2) ->
	  mintype(typeOfVar m1 v, typeOfVar m2 v)
      | Crypt(m1, Var(vv)) when v=vv -> SKey
      | Publ(Var(vv)) when v=vv -> Name
      | Priv(Var(vv)) when v=vv -> Name
      | Crypt(m1, Publ(Var(vv))) when vv=v -> Name
      | Crypt(m1, Priv(Var(vv))) when vv=v -> Name
      | Crypt(m1, kk) ->
	  mintype(typeOfVar m1 v, typeOfVar kk v)
      | _ -> Gen


	   
      
