(*
	Pl_Parser.ml
	14/03/2003
	Giacomo Baldi

	Protocol logic parser
*)

open Tokenizer
open Message
open Logic

(*
	This parser is used to extract a formula from a text file given the following sintax:

	<Logic> ::= <Formula>;
	<Formula> ::=
		|| <Variable> = <Message>
		|| <Variable> <> <Message>
		|| :> <Message>
		|| !:> <Message>
		|| forall <index>.<Name>:
		|| exists <index>.<Name>:
		|| not <Formula>
		|| ! <Formula>
		|| <Formula> and <Formula>
		|| <Formula> & <Formula>
		|| <Formula> or <Formula>
		|| <Formula> | <Formula>
		|| <Formula> implies <Formula>
		|| <Formula> -> <Formula>
		|| (<Formula>)

	<Message>::= (<Message>, <Message>)
		|| {<Message>}<Message>
		|| <Key>
		|| <PName>
		|| <Pname>+
		|| <Pname>-
		|| <Nonce>
		|| <Variable>
		|| <Variable>+
		|| <Variable>-

	<Name>::= an uppercase string representing the name of the principal
	<Key>::= a string beginning with 'k'
	<PName>::= an uppercase string representin principal names or keys if followed by (+,-)
	<Nonce>::= a string beginning with 'n'
	<Variable>::= a lowercase string not beginning with 'k' or 'n'

	- Comments are allowed everywhere. A comment can be a '//' and means that
	everything on that line is ignored. A multiline comment begin with '/*'
	and end with '*/'. Everything in between is ignored.

	precedences between operator are:
	forall, exists >  implies > and, or > not, derive, equal
*)


(*
	Internally used.
	- Pexp represent an expression already formed
	- PNot, PAnd, POr, Impl represent the boolean operators ! & | ->
	- Plp stands for an open par '('
	- Der is a derivation of message msg
	- Eq is equality between var and msg
	- Fa, Ex are universally and existentially quantifiers with references to index and principal name
	- Token is used to mark a parsed message
	- End is end of parsing
*)

type pars =
	| Pexp of pls
	| PNot
	| PAnd
	| POr
	| Impl
	| DImpl
	| Fa of string*string
	| Ex of string*string
	| Plp
	| Token of string
	| End

(*
	ParseError is raised when a general error occurs
	CantReduce is raised when trying to reduce expression on the stack
*)

exception ParseError
exception CantReduce

(*
	error codes.
*)

let errOK							= 0;;
let errCOMMENT_NOT_OPENED			= 1;;
let errCOMMENT_NOT_CLOSED			= 2;;
let errUNKNOWN_TOKEN				= 3;;
let errMISSING_OPENING_ROUND_PAR	= 4;;
let errMISSING_CLOSING_ROUND_PAR	= 5;;
let errMISSING_SEMICOLON			= 6;;
let errMISSING_COLON				= 7;;
let errMISSING_DOT					= 8;;
let errEXPECTED_IDE					= 9;;
let errUNEXPECTED_TOKEN				= 10;;
let errWRONG_KEY					= 11;;
let errWRONG_VAR					= 12;;
let errMALFORMED_EXPR				= 13;;
let errUPPERCASE_PRINCIPAL			= 14;;
let errMISSING_VARIABLE				= 15;;
let errMISSING_COMMA				= 16;;
let errMISSING_OPENING_CURLY_PAR	= 17;;
let errMISSING_CLOSING_CURLY_PAR	= 18;;
let errMISSING_INDEX				= 19;;


(*
	error string table.
*)

let errorString = Array.create 20 "";;
errorString.(errOK) <- "Ok.";;
errorString.(errCOMMENT_NOT_OPENED) <- "Comment not opened";;
errorString.(errCOMMENT_NOT_CLOSED) <- "Comment not closed";;
errorString.(errUNKNOWN_TOKEN) <- "Unknown token";;
errorString.(errMISSING_OPENING_ROUND_PAR) <- "Missing (";;
errorString.(errMISSING_CLOSING_ROUND_PAR) <- "Missing )";;
errorString.(errMISSING_OPENING_CURLY_PAR) <- "Missing {";;
errorString.(errMISSING_CLOSING_CURLY_PAR) <- "Missing }";;
errorString.(errMISSING_SEMICOLON) <- "Missing ;";;
errorString.(errMISSING_COLON) <- "Missing :";;
errorString.(errMISSING_DOT) <- "Missing .";;
errorString.(errMISSING_COMMA) <- "Missing ,";;
errorString.(errEXPECTED_IDE) <- "Identifier expected";;
errorString.(errUNEXPECTED_TOKEN) <- "Unexpected token";;
errorString.(errWRONG_KEY) <- "Wrong key";;
errorString.(errWRONG_VAR) <- "Wrong variable";;
errorString.(errMALFORMED_EXPR) <- "Malformed expression";;
errorString.(errUPPERCASE_PRINCIPAL) <- "Principal name should be uppercase";;
errorString.(errMISSING_VARIABLE) <- "Missing variable in equality";;
errorString.(errMISSING_INDEX) <- "Missing index";;



(*
	returns a string representing the line and the column of the error position
*)
let whereError =
  function () -> " ("^(string_of_int(fst(where())))^","^(string_of_int(snd(where())))^")"


let errorCode = ref(0)
let errorData = ref ("")
let errorMsg =
  function () -> "ER"^string_of_int(!errorCode)^": "^errorString.(!errorCode)^" "^(!errorData)^" "^whereError()^"\n"


(*
   used to signal an error while parsing
 *)
let putError =
  function (code, s) -> errorData:= s; errorCode:=code

let getErrorMsg = function () -> errorMsg()
let getErrorCode = function () -> !errorCode


(*
	check if a string is uppercase
*)
let isUpper =
  function s ->
    let isUp =
      function ch ->
	match ch with
	| 'A'..'Z' -> true
	| '_' -> true
	| _ ->  false
    in
    let rec isU =
      function (s, i, m) ->
	if (i<m) then
	  if isUp(s.[i]) then isU(s, i+1, m)
	  else false
	else true
    in isU(s,0, String.length s)


(*
   reads from tokenizer until current comment is closed
 *)
let rec ignoreComment =
  function () ->
    let a = getToken() in
    match a with
    | LSym("*") ->
	let b = getToken() in
	if b=LSym("/") then ()
	else (pushback(b); ignoreComment())
    | LStop -> (raise ParseError)
    | _ -> ignoreComment()


(*
   extracts current token taking care of comments
 *)
let rec readToken =
  function () ->
    let a = getToken() in
    match a with
    | LSym("/") ->
	let b = getToken() in
	if b=LSym("/") then (flushline(); readToken())
	else
	  if b=LSym("*") then ((let ss = whereError() in try (ignoreComment()) with ParseError -> putError(errCOMMENT_NOT_CLOSED,ss)); readToken())
	  else (pushback(b); a)
    | LSym("*") ->
	let b = getToken() in
	if b=LSym("/") then (putError(errCOMMENT_NOT_OPENED,""); raise ParseError)
	else (pushback(b); a)
    | LUnknown(a) -> (putError(errUNKNOWN_TOKEN,""); raise ParseError)
    | _ -> a

(*
   This is a bunch of functions to parse messages.
   We parse the message with getMessage.

   While parsing with getMessage we scan the input for :
   - '(' which means we expect a couple
   - '{' which means we expect a crypt
   - '?' which means we expect a binding var
   - identifier, which means we expect something as a variable or a key.

   To parse a couple we first parse the first message, then we search for a ','
   then we parse the second message expecting a final matching ')'.

   To parse a crypt we first parse the message, then we search for '}' then we parse
   the key.

   getVal is used to recognize variables, names, nonces and keys.
   If we have an uppercase identifier it is a principal name or an asymmetric key
   if it is followed by '+' or '-'.
   If we have an identifier starting with 'k' it is a key.
   If we have an identifier starting with 'n' it is a nonce.
   Every other identifier is a variable which can't be decorated

   getName parse <name>: in a quantifier scope
   getIndex parse <index>. in a quantifier scope
   getIndexedDatum extracts name and index information from tokens
 *)


let getName =
  function () ->
    let a = readToken() in
    let b = readToken() in
    match a,b with
    | LIde(n), LSym(".") when isUpper(n) -> n
    | LIde(n), LSym(".") -> (putError(errUPPERCASE_PRINCIPAL,""); raise ParseError)
    | LIde(n), _ -> (putError(errMISSING_DOT,""); raise ParseError)
    | _ -> (putError(errUNEXPECTED_TOKEN,""); raise ParseError)

let getIndex =
  function () ->
    let a = readToken() in
    match a with
    | LIde(n) ->
	let b = readToken() in
	if b<>LSym(":") then (putError(errMISSING_COLON,""); raise ParseError)
	else n
    | _ -> (putError(errUNEXPECTED_TOKEN,""); raise ParseError);;


let getIndexedDatum =
  let getNum =
    function st -> try (int_of_string(st)) with _ -> -1
  in
  let splitName =
    function ws ->
      let sl = String.length ws in
      let pl = (try ((String.rindex ws '_' )) with Not_found -> -1) in
      if (pl = -1 || pl=sl-1) then (putError(errMISSING_INDEX,""); raise ParseError)
      else (pl, String.sub ws 0 pl, String.sub ws (pl+1) (sl-pl-1))
  in
  function s ->
    let nl = String.length s in
    let (pos, n, i) = splitName(s) in
    let im = String.sub s (pos+1) (nl-pos-2) in
    if (isUpper(n)) then (
      match s.[nl-1] with
      | '+' ->(
	  let vi = getNum(im) in
	  if vi<> -1 then PNp(n,vi)
	  else PNp(n^im,0)
	 )
      | '-' ->(
	  let vi = getNum(im) in
	  if vi<> -1 then PNm(n,vi)
	  else PNm(n^im,0)
	 )
      | _ ->(
	  let vi = getNum(i) in
	  if vi<> -1 then PN(n,vi)
	  else PN(n^i,0)
	 )
     )
    else (
      if (s.[0]<>'n' && s.[0]<>'k') then (
	match s.[nl-1] with
	| '+' ->(
	    let vi = getNum(im) in
	    if vi<> -1 then (
	      let (x, nn, np) = splitName(n) in
	      Publ(Var(nn,(np,vi)))
	     )
	    else Publ(Var(n,(im,0)))
	   )
	| '-' ->(
	    let vi = getNum(im) in
	    if vi<> -1 then (
	      let (x, nn, np) = splitName(n) in
	      Priv(Var(nn,(np,vi)))
	     )
	    else Priv(Var(n,(im,0)))
	   )
	| _ ->(
	    let vi = getNum(i) in
	    if vi<> -1 then (
	      let (x, nn, np) = splitName(n) in
	      Var(nn,(np,vi))
	     )
	    else Var(n,(i,0))
	   )
       )
      else (
	let vi = getNum(i) in
	if vi<> -1 then (
	  if (s.[0]='k') then K(n,vi) else NO(n,vi)
	 )
	else (
	  if (s.[0]='k') then K("_"^i^"_"^n,0) else NO("_"^i^"_"^n,0)
	 )
       )
     )


let getVar =
  function s ->
    match getIndexedDatum(s) with
    | Var(p) -> Var(p)
    | PN(p,i) -> PN(p,i)
    | _ -> (putError(errWRONG_VAR,""); raise ParseError)


let rec getMessage =
  function() ->
    let a = readToken() in
    match a with
    | LSym("(") -> getCouple()
    | LSym("{") -> getCrypt()
    | LIde(v) -> (pushback(a); getVal())
    | _ -> (putError(errUNEXPECTED_TOKEN,""); raise ParseError)
and getCouple =
  function () ->
    let m1 = getMessage() in
    let v1 = readToken() in
    if v1<>LSym(",") then (putError(errMISSING_COMMA,""); raise ParseError)
    else
      let m2 = getMessage() in
      let v2 = readToken() in
      if v2<>LSym(")") then (putError(errMISSING_CLOSING_ROUND_PAR,""); raise ParseError)
      else Couple(m1,m2)
and getCrypt =
  function () ->
    let m1 = getMessage() in
    let v1 = readToken() in
    if v1<>LSym("}") then (putError(errMISSING_CLOSING_CURLY_PAR,""); raise ParseError)
    else
      let k1 = getVal() in Crypt(m1,k1)

and getVal =
  function () ->
    let a = readToken() in
    match a with
    | LIde(s) -> (
	let k = readToken() in
	match k with
	| LSym("+") -> getIndexedDatum(s^"+")
	| LSym("-") -> getIndexedDatum(s^"-")
	| _ -> (pushback(k); getIndexedDatum(s))
       )
    | _ -> (putError(errEXPECTED_IDE,""); raise ParseError);;



(*
   while putting things on the parsing stack with parseExp, we use reducer
   trying to pack the current stack.
 *)


let pvalue =
  function p ->
    match p with
    | PNot -> 0
    | PAnd
    | POr -> 1
    | Impl -> 2
    | DImpl -> 2
    | Fa(i,n) -> 3
    | Ex(i,n) -> 3
    | _ -> -1



let reducer =
  function (stack, pr) ->
    match stack with
    | Pexp(e)::PNot::st -> Pexp(Not(e))::st
    | Pexp(e1)::PAnd::Pexp(e2)::st when pvalue(pr)>=pvalue(PAnd) -> Pexp(And(e2,e1))::st
    | Pexp(e1)::POr::Pexp(e2)::st when pvalue(pr)>=pvalue(POr) -> Pexp(Or(e2,e1))::st
    | Pexp(e1)::Impl::Pexp(e2)::st when pvalue(pr)>=pvalue(Impl) -> Pexp(Or(Not(e2),e1))::st
    | Pexp(e1)::DImpl::Pexp(e2)::st when pvalue(pr)>=pvalue(DImpl) -> Pexp(Or(And(Not(e2),Not(e1)),And(e1,e2)))::st
    | Pexp(e)::Fa(i,n)::st when pvalue(pr)>=pvalue(Fa(i,n)) -> Pexp(Forall(i,n,e))::st
    | Pexp(e)::Ex(i,n)::st when pvalue(pr)>=pvalue(Ex(i,n)) -> Pexp(Exists(i,n,e))::st
    | _ -> raise CantReduce;;

(*
   This is the heart of process parsing.
   Given a token and a stack (initially empty) we analyze
   the token and
   ...
   next we'll call reduce_all for the last reduction
 *)
let rec parseExp =
  function (tok, stack) ->
    match (tok, stack) with
    | LIde("true"), _ -> Pexp(True)::stack
    | LIde("false"), _ -> Pexp(False)::stack
    | LIde("forall"), _ ->
	let n = getName() in
	let i = getIndex() in Fa(i,n)::stack
    | LIde("exists"), _ ->
	let n = getName() in
	let i = getIndex() in Ex(i,n)::stack
    | LSym("&"), _
    | LIde("and"), _ -> (try parseExp (tok, reducer(stack, PAnd)) with CantReduce -> PAnd::stack)
    | LSym("|"), _
    | LIde("or"), _ -> (try parseExp (tok, reducer(stack, POr)) with CantReduce -> POr::stack)
    | LSym("=>"), _
    | LIde("implies"), _ -> (try parseExp (tok, reducer(stack, Impl)) with CantReduce -> Impl::stack)
    | LSym("<=>"), _
    | LIde("iff"), _ -> (try parseExp (tok, reducer(stack, DImpl)) with CantReduce -> DImpl::stack)
    | LSym("!"), _
    | LIde("not"), _ -> PNot::stack
    | LSym(":>"), _
    | LIde("derive"), _ -> let ms = getMessage() in Pexp(Derive(ms))::stack
    | LSym("="), Token(s)::st
    | LIde("equal"), Token(s)::st -> let ms = getMessage() in Pexp(Equal(getVar(s),ms))::st
    | LSym("<>"), Token(s)::st -> let ms = getMessage() in Pexp(Diff(getVar(s),ms))::st

    | LIde(s), _ ->  Token(s)::stack

    | LSym("("), _ -> Plp::stack
    | LSym(")"), Pexp(e)::Plp::st -> Pexp(e)::st
    | LSym(")"), _ -> parseExp (tok, reducer(stack, Fa("","")))
    | LSym(";"), _ -> End::stack
    | LStop, _ -> (putError(errMISSING_SEMICOLON,""); raise ParseError)
    | _ -> putError(errUNKNOWN_TOKEN,""); raise ParseError;;


let rec reduce_all =
	function stack ->
		match stack with
			| [Pexp(e)] -> e
			| [] -> (putError(errMISSING_SEMICOLON,""); raise ParseError)
			| _ -> reduce_all ( reducer (stack, Fa("","")));;


(*
	Calls parseExp for each token until end of input or end of parsing. At this point we return
	the current stack without the final End.
*)
let rec parser =
    function stack ->
      let a = readToken() in
      let d =
	(match a with
	| LSym("=") ->
	    (let b = readToken() in
	    if b<>LSym(">") then (pushback(b); a)
	    else LSym("=>"))
	| LSym(":") ->
	    (let b = readToken() in
	    if b<>LSym(">") then (pushback(b); a)
	    else LSym(":>"))
	| LSym("<") ->
	    (let b = readToken() in
	    if (b=LSym(">")) then LSym("<>")
	    else if (b<>LSym("=")) then (pushback(b); a)
	    else (
	      let c = readToken() in
	      if (c=LSym(">")) then LSym("<=>")
	      else (pushback(c); LSym("<="))
	     )
	    )
	| _ -> a) in
      let ss = parseExp (d, stack) in
      let c = List.hd ss in
      if c=End then stack
      else parser ss;;


let parseFormula =
  function () ->
    let (f,ff) =
      try  (true, reduce_all (parser [])) with
      | CantReduce -> (putError(errMALFORMED_EXPR,""); (false,False))
      | ParseError -> (false, False)

    in
    if f then ff
    else False


let rec parseKb =
  function () ->
    let m = getMessage() in
    let a = readToken() in
    match a with
    | LSym(",") -> m::(parseKb())
    | LSym(";") -> [m]
    | _ -> (putError(errMISSING_SEMICOLON,""); raise ParseError)
(*
   Starts to read from fname.
   Close fname.
   Theese are used in setup and shutdown phase.
 *)
let init_parser =
	function fname -> init_tokenizer(fname);;
let done_parser =
	function () -> done_tokenizer();;

