(*
	Message.mli
	23/01/2003
	Giacomo Baldi

	Definition of operations over messages and variables
*)
(**
	This module contains the implementation of the operations over messages and variables.
	This is the fundamental module used by the search engine and by all data structures.
*)

(**
	A value of type pname is a couple which represents the name of a principal and its index
*)
type pname = string*int

(**
	A variable has a name and belongs to a principal.
*)
type variable = string*pname

(**
	When a variable becomes a symbolic one, it acquires a type which restrict the type
	of elements belonging to the set of messages it represents.
*)
type svtype =
	| Gen (** Everything that can be derived from the knowledge*)
 	| GenR (** Everything that is in the knowledge*)
	| SKey (** a symmetric key *)
	| Name (** a name of a principal *)
	| RName (** a name of a principal of which we know asymmetric keys *)
	| PMKey (** a public key of which we know the private one *)
	| MPKey (** a private key of which we know the public one *)


(**
	A message can be:
	- a symmetric key
	- a name of a principal
	- a public key of a principal
	- a private key of a principal
	- a nonce
	- a variable
	- a binding variable
	- a symbolic variable
	- a cryptogram
	- a couple
	- a decorated name

	* a symbolic variable is a variable which represents a subset of the message derivable
	from the knowledge it is associated with.
	
	* a decorated message is a message followed by  a + or a -
*)
type msg =
	| K of string*int (** a symmetric key *)
	| PN of string*int (** a principal name  *)
	| PNp of string*int (** a private key *)
	| PNm of string*int (** a public key *)
	| NO of string*int  (** a nonce *)

	| Var of variable (** a variable *)
	| BVar of variable (** a binding variable *)
	| SVar of variable*int*svtype (** a symbolic variable (variable, knowledge id, type *)

	| Crypt of msg*msg  (** a cryptogram *)
	| Couple of msg*msg (** a couple *)
	| Publ of msg (** a message decorated with + *)
	| Priv of msg (** a message decorated with - *)
	| Bottom  (** void message *)
	| Trick of (msg*msg) list  (** a message internally used in assignment to speed up things *)



(**
	setVariable takes a pname and a variable.
	@return a variable belonging to the principal specified by the pname value
*)
val setVariable: pname*variable -> variable

(**
	isKey takes a message m.
	@return true if m can be used as a key in a cryptogram, false otherwise.
*)
val isKey: msg -> bool

(**
	isGroundKey takes a message m.
	@return true if m can be used as a key in a cryptogram and it isn't a variable, false otherwise.
*)
val isGroundKey: msg -> bool

(**
	checkmsg takes a message m.
	@return true if m is a semantically valid message, false otherwise.
*)
val checkmsg: msg -> bool

(**
	hasntFreeVar takes a message m.
	@return true if m hasn't a variable which is not binding, false otherwise.
*)
val hasntFreeVar: msg -> bool


(**
	compl takes a message m.
	@return the complement of a key. If m isn't a key return Bottom.
*)
val compl: msg -> msg

(**
   typeOfVar takes a message m and a variable v.
   @return the type of variable v in m.
*)
val typeOfVar: msg -> variable -> svtype

(**
	notSVar takes a message m.
	@return true in m isn't  a symbolic variable, false otherwise.
*)
val notSVar: msg -> bool

(**
	notVar takes a message m.
	@return true in m isn't  a  variable, false otherwise.
*)
val notVar: msg -> bool

(**
	tostring takes a message m.
	@return the string representation of m.
*)
val tostring: msg -> string

(**
	tostringVarList takes a variable  list vl.
	@return the string representation of vl.
*)
val tostringVarList: variable list -> string

(**
	toName takes a message m.
	@return Bottom if m isn't an asymmetric key, the name of the principal otherwise.
*)
val toName: msg -> msg

(**
	toPNp takes a message m.
	@return Bottom if m isn't a name, the public key of the principal otherwise.	
*)
val toPNp: msg -> msg

(**
	toPNm takes a message m.
	@return Bottom if m isn't a name, the private key of the principal otherwise.	
*)
val toPNm: msg -> msg

(**
	mintype takes a couple of types.
	used internally.
*)
val mintype: svtype*svtype -> svtype

(**
	comparable takes a couple of types.
	used internally.
*)
val comparable: svtype*svtype -> bool

(**
	ispkey takes a message m.
	@return true if m is a public key, false otherwise.
*)
val ispkey: msg -> bool

(**
	ismkey takes a message m.
	@return true if m is a private key, false otherwise.
*)
val ismkey: msg -> bool

(**
	isskey takes a message m.
	@return true if m is a symmetric key, false otherwise.
*)
val isskey: msg -> bool

(**
	isname takes a message m.
	@return true if m is a name, false otherwise.
*)
val isname: msg -> bool

(**
	iscrypt takes a message m.
	@return true if m is a cryptogram, false otherwise.
*)
val iscrypt: msg -> bool

(**
	isBinding takes a couple of messages m1, m2.
	@return true if m is a cryptogram, false otherwise.
*)
val isBinding: msg*msg -> bool


(**
	prntMsg takes a message m.
	print out a representation of m.
*)
val prntMsg: msg -> unit

(**
	prntVar takes a variable v.
	print out a representation of v.
*)
val prntVar: variable -> unit

(**
	prntMsgHtml takes a message m.
	print out a representation of m.
*)
val prntMsgHtml: msg -> unit

(**
	prntVarHtml takes a variable v.
	print out a representation of v.
*)
val prntVarHtml: variable -> unit

(**
	prntVar takes a variable list vl.
	print out a representation of vl.
*)
val prntVarlist: variable list -> unit

(**
	prntVar takes a variable list vl.
	print out a representation of vl.
*)
val prntVarlistHtml: variable list -> unit
