(*
   tracer.ml
   30/08/2003
   Giacomo Baldi

   implementation of search in space state
*)

open Message
open Principal
open Kmanager
open Node
open Parsecmdline
open Logic
open Step
open Utils
open Thestack
open Action
open Verifier
open Csolver
open Join
open Prenex
open Evalh
open Misc
open H_var

exception StopSearch
exception AttackFound
exception NextTrace
    
type ss = {
    mutable bfm: float;
    mutable maxbf: int;
    mutable minbf: int;
    mutable len_traces: int list;
    mutable son_traces: float list;
    
    mutable initial_nodes: int;
    mutable current_node: int;
    mutable closed_nodes: int;
    mutable transitions: int;
    mutable num_nodes: int;
    mutable num_traces: int;
    mutable num_terminal_traces: int;
    mutable num_attacks: int;

    mutable root_node: node;
    mutable startnodes: node list;
    mutable startjoins: (node * cs)  list list;
    mutable startpls: npls list;
    mutable joinpls: npls list;

    mutable kman: km;
    mutable trans:  node -> node;
    mutable init:  node -> node;
    mutable join: node -> (node) list;
    mutable opts: options;

    mutable jointime: float;
    mutable searchtime: float;
    mutable verifytime: float;
    mutable elapsed: float;

    mutable vflag: bool;	
  }


let init_search =
  function prns ->   
    function kbl -> 
      function options -> 
	let ct = Context.void() in
	let asg = Assignment.void() in
	let kmanage = Kmanager.initKM (Sys.max_array_length) in
	let kb1 = (Kmanager.newK kmanage) in
	let pool = Princpool.create (prns) options.max_princ in
	let transition = getTransitionFunction kmanage in
	let initjoinfn = initjoin pool kmanage options.ic in
	let joinmodel = joincontext kmanage options.greed in
	let kb = (Kmanager.expandK(kmanage, kb1, asg,(   fold2 kbl (function a -> function b -> Couple(a,b))  )))  in
	let st = Status.create(ct,asg,kb) in
	let root = Node.create(st) in
	let current_status = {
	  bfm = 0.0;
	  maxbf = 0;
	  minbf = 0;
	  len_traces = [];
	  son_traces = [];
	  initial_nodes = getinitialnodes (len prns) (options.max_princ);
	  current_node = 0;
	  closed_nodes = 0;
	  transitions = 0;
	  num_nodes = 0;
	  num_traces = 0;
	  num_terminal_traces = 0;
	  num_attacks = 0;
	  root_node = root;
	  startnodes = [];
	  startjoins = [];
	  startpls = [];
	  joinpls = [];
	  kman = kmanage;
	  trans = transition;
	  init = initjoinfn;
	  join = joinmodel;
	  opts = options;
	  jointime = 0.0;
	  searchtime = 0.0;
	  verifytime = 0.0;
	  elapsed = 0.0;
	  vflag = false;
	} in current_status

let format_time =
  function time ->
    let hour = int_of_float(ceil time) / 3600 in
    let min = (int_of_float(ceil time) - (hour * 3600)) / 60 in
    let sec = (int_of_float(ceil time) - (3600*hour + 60*min)) in
    if (hour < 0) then ("infinity")
    else (string_of_int(hour)^":"^string_of_int(min)^":"^string_of_int(sec))


let rec prntTrace =
  function ss ->
    function nn ->
      if Node.isNul(nn) then  print_string("\n-------\n Trace ("^string_of_int(ss.num_traces)^")|<<---------------------------|\n-------\n\n")
      else
	let np = Node.getfather nn in
	(
	 prntTrace ss (np);
	 Node.prnt nn
	)

let rec build_intruder =
  function nn ->
    function aaaa ->
      let ff = Node.getfather nn in
      if Node.isNul(ff) then print_string("Intruder:\n")
      else (
	build_intruder ff aaaa;
	let acc = Node.getlabel nn in
	match acc with
	| Output((nm,ii),mm,s) -> print_string(nm^"_"^string_of_int(ii)^" -> I: ");prntMsg (aaaa mm); print_string("\n") 
	| Input((nm,ii),mm,s) -> print_string("I -> "^nm^"_"^string_of_int(ii)^": ");prntMsg (aaaa mm); print_string("\n")
	| _ -> ()
)

let rec build_intruder_html =
  function nn ->
    function aaaa ->
      let ff = Node.getfather nn in
      if Node.isNul(ff) then () 
      else (
	build_intruder_html ff aaaa;
	let acc = Node.getlabel nn in
	match acc with
	| Output((nm,ii),mm,s) -> 
	    print_string(nm^"<sub>"^string_of_int(ii)^"</sub> -&gt <b>I</b>: ");
	    prntMsgHtml (aaaa mm); 
	    print_string("<br>") 
	| Input((nm,ii),mm,s) -> 
	    print_string("<b>I</b> -> "^nm^"<sub>"^string_of_int(ii)^"</sub>: ");
	    prntMsgHtml (aaaa mm); 
	    print_string("<br>")
	| _ -> ()
)

let calcexpect =
  function ss ->
    let i = ref(-1) in
    let vals = map (ss.len_traces) (function lt -> i:=!i+1; if !i < ss.current_node then 0.0 else  (select ss.son_traces !i) *. (ss.bfm ** (float_of_int(lt) ))) in
    fold vals (function en -> function ctot -> en +. ctot) (float_of_int(ss.num_nodes))


let calcbf =
  function ss ->
    function ns ->
      if ns>(ss.maxbf) then ss.maxbf <- ns else ();
      if ns<(ss.minbf)||(ss.minbf=0) then ss.minbf <- ns else ();
      ss.transitions <- ss.transitions+ns;
      ss.closed_nodes <- ss.closed_nodes+1;
      ss.bfm <- (float_of_int(ss.transitions)) /. (float_of_int(ss.closed_nodes))

let feedback = 
  function ss -> 
    if (ss.opts.feedback) then( 
      let expect = calcexpect ss in 
      let perc = (100.0 *. (float_of_int(ss.num_nodes) /. expect)) in 
      let tm = Sys.time() -. ss.searchtime in 
      let time = ((tm *. 100.0) /. perc) -. tm in let remaining = format_time time in 
      let elapsed = format_time tm in
      prerr_string("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
      prerr_string(string_of_int(int_of_float(perc))^"%"); 
      prerr_string(" elapsed "^elapsed^" "); 
      prerr_string("~ remaining "^remaining^" ");
      prerr_string(" subtree "^string_of_int(ss.current_node+1)^"/"^string_of_int(ss.initial_nodes));
      prerr_string(" ("^string_of_int(ss.num_attacks)^") "); 
      if (ss.vflag) then (
	ss.vflag <- false; 
	prerr_string(" - V ")
       ) else (
	prerr_string(" - S ")
       ); 
(*prerr_string(string_of_float(!bfm));*) 
      flush stderr 
     ) else ()


let printAttack = 
  function ss -> 
    function nn -> 
      function conj ->
	function nid -> 
	  function cssf -> 
	    if (ss.opts.attack_type = 2) then(
(*html*) 
	      ss.num_attacks <- ss.num_attacks+1;
	      print_string("<center><table border=1><tr><td colspan=2 >"); 
	      print_string(" <b>ATTACK "^string_of_int(ss.num_attacks)^" -- (Id: "^string_of_int(nid)^")</b>");
	      print_string("</td></tr><tr><td><b>Violated Constraints</b></td><td><b>Open Variables</b></td></tr><tr><td>"); 
	      let (cti,asgi,ki) = Node.getstatus (select ss.startnodes ss.current_node) in 
	      let (ctx,asgx,kx) = Node.getstatus nn in ( 
	      Logic.prntnplHtml ([conj]);
	      print_string("</td><td>"); 
	      let asgcs = Csolver.toasg cssf in 
	      let mdls = Csolver.linearize cssf in 
	      let cpl1 = couplify asgcs mdls in 
	      let (ax,cx) = head cpl1 in 
	      let newasg = (Assignment.comp ax asgx) in 
	      let (nv,vars) = Context.getOpenVariables cti in 
	      print_list vars (function v -> Message.prntMsgHtml v; print_string(" -&gt "); Message.prntMsgHtml (newasg v);print_string("<br>"));
	      print_string("</td></tr><tr><td><b>Knowledge</b></td><td><b>Context</b></td></tr><tr><td>");
	      Knowledge.prntHtml (Knowledge.applyall kx newasg);
	      print_string("</td><td>"); 
	      Context.prntHtml cti newasg;
	      print_string("</td></tr>"); 
	      print_string("<tr><td><b>Model</b></td><td><b>Intruder process<b></td></tr><tr><td>");
	      Csolver.prntHtml cx;
	      print_string("</td><td>");
	      build_intruder_html nn newasg;
	      print_string("</td></tr></table></center><br><hr><br>")
	       ) 
	     ) else ( 
(*ascii*)
	      ss.num_attacks <- ss.num_attacks+1; 
	      print_string(" ATTACK "^string_of_int(ss.num_attacks)^" -- (Id: "^string_of_int(nid)^")");
	      print_string("\n--------------------------------------------------------------\n\n");
	      let (cti,asgi,ki) = Node.getstatus (select ss.startnodes ss.current_node) in 
	      let (ctx,asgx,kx) = Node.getstatus nn in (
	      print_string("Violated Constraints:\n"); 
	      Logic.prntnpl ([conj]); 
	      let asgcs = Csolver.toasg cssf in 
	      let mdls = Csolver.linearize cssf in 
	      let cpl1 = couplify asgcs mdls in 
	      let (ax,cx) = head cpl1 in 
	      let newasg = (Assignment.comp ax asgx) in 
	      print_string("\n***************\n");
	      print_string("Open Variables:\n"); 
	      let (nv,vars) = Context.getOpenVariables cti in 
	      print_list vars (function v -> Message.prntMsg v; print_string(" -> "); Message.prntMsg (asgx v);print_string("\n")); print_string("Context:\n"); 
	      Context.prnt cti newasg; 
	      print_string("Knowledge:\n"); 
	      Knowledge.prnt (Knowledge.applyall kx newasg); 
	      print_string("\nModel:\n");
	      Csolver.prnt cx;
	      build_intruder nn newasg;
	      print_string("\n\n\n");
	      print_string("There are "^string_of_int(len cpl1)^" possible models.");
 	      print_string("\n\n"); 
	      if (ss.opts.mga) then ()
	      else (
		print_list (tail cpl1) (
		function (aa,cc) ->
		  print_string("\nModel:\n");
		  Csolver.prnt cc;		
	       )
	       )
	       );
	      print_string("\n***************\n");
	      print_string("\n--------------------------------------------------------------\n\n");

	     );
	    flush stdout
	      



let verification =
  function ss ->
    function nn ->
      function cf ->
	let ask = 
	  function () ->
	    if (ss.opts.interactive) then (
	      prerr_string("\nAttack found ("^string_of_int(ss.num_attacks)^"). [N]ext, [S]top, [O]ther trace? ");
	      flush stderr;
	      let res = try (input_line stdin) with End_of_file -> "Y" in
	      if (res="O") then (raise NextTrace)
	      else if (res="S") then (raise StopSearch)
	      else ()
	     ) else ()
	in	
(*	if (ss.opts.lazyness) then ( *)
	(* lazy verification *)
	ss.vflag <- true; 
	feedback ss;

	let conjuncts = select ss.startpls ss.current_node in
	let jpl = select ss.joinpls ss.current_node in
	try (
	  iter conjuncts (
	  function ff -> 	    
	    ss.vflag <- true; 
	    feedback ss;
	    let (conj,verf) = fast_checkpl ([ff]) (cf) (ss.kman) nn in
	    try (
	      iter verf (
	      function (pl,asg,ccf) ->
		let (ctn,asgn, kbn) = Node.getstatus nn in
		let newasg = Assignment.comp asgn asg in
		let newct = Context.applyall ctn newasg in
		let newnode = Node.create((newct,newasg,kbn)) in
		Node.setfather (newnode, nn);
		Node.setlabel (newnode, Join(("Lazy Check",Node.getid newnode),newasg));
		ss.vflag <- true; 
		feedback ss;
		if (fast_sat_model (jpl) (ccf) (ss.kman) (newnode)) then (		
		  printAttack ss newnode ff (Node.getid nn) ccf; 
		  raise AttackFound
		 )
		else ()
	     )
	     ) with AttackFound -> ask(); if (ss.opts.mga && (not ss.opts.interactive)) then raise NextTrace else ();
	 )
	 ) with NextTrace -> ()	      

(*
   )
   else ( *)
(* eager verification *)
(*
   let conjuncts = select ss.startpls ss.current_node in
   try (
   iter conjuncts (
   function ff -> 	    
   ss.vflag <- true; 
   feedback ss;
   let (conj,verf) = fast_checkpl ([ff]) (cf) (ss.kman) nn in
   iter verf (
   function (pl,asg,ccf) ->
   let (ctn,asgn, kbn) = Node.getstatus nn in
   let newasg = Assignment.comp asgn asg in
   let newct = Context.applyall ctn newasg in
   let newnode = Node.create((newct,newasg,kbn)) in
   Node.setfather (newnode, nn);
   Node.setlabel (newnode, Join(("Eager Check",Node.getid newnode),newasg));
   printAttack ss newnode ff (Node.getid nn) ccf; 
   );
   ask(); 
   if (ss.opts.mga && (not ss.opts.interactive)) then raise NextTrace else ();
   )
   ) with NextTrace -> ();	      

   )
 *)

let iterstack =
  function ss ->
    function cf ->
      function nstack ->
	let cti = Sys.time() in
	let top = Thestack.top nstack in
	let nk = ss.trans top in
	ss.num_nodes <- (if(Node.getid nk > ss.num_nodes) then Node.getid nk else ss.num_nodes);
	if (cti -. ss.elapsed > 1.0 ) then (
	  ss.elapsed <- cti;
	  feedback ss
	 )
	else ();
	if (Node.isNul nk) then (
	  let nss = Thestack.pop nstack in
	  if (Node.isLeaf top) then (
	    ss.num_traces <- ss.num_traces+1;
	    if (Node.isTerminal top || ss.opts.trace_terminal) then (
	      ss.num_terminal_traces <- ss.num_terminal_traces +1;
	      (*print_string(string_of_int(Node.getid top)^"\n"); flush stdout;*)
	      if (ss.opts.output_type = 2 ) then (
		let stime = Sys.time() in
		(verification ss top cf);
		ss.verifytime <- ss.verifytime +. Sys.time() -. stime;
	       )
	      else
		if (ss.opts.output_type = 0) then prntTrace ss top 
		else
		  if (ss.opts.output_type = 1 && (not ss.opts.lazyness)) then Gviz.prntTrace top (ss.opts.go) 
		  else 
		    if (ss.opts.output_type = 1) then (
		      (*
			 let mdls = getmodels (select ss.joinpls ss.current_node) (ss.kman) top in
			 let nodes = ss.join top mdls in
			 iter nodes (function (n1,cf1) -> Gviz.prntTrace n1 (ss.opts.go)); 		       
		       *)Gviz.prntTrace top (ss.opts.go) 
		     )
		    else ();	    
	      nss
	     ) 
	    else (
	      if (ss.opts.output_type=0 && (not ss.opts.trace_terminal)) then prntTrace ss top else ();
	      if (ss.opts.output_type=1 && (not ss.opts.trace_terminal)) then Gviz.prntTrace top (ss.opts.go) else ();	    
	      nss
	     )
	   )
	  else (
	    calcbf ss (Node.getSons top);
	    nss
	   )
	 )
	else (
	  Thestack.push nstack nk
	 )



let do_joins =
  function ss ->
    function pl ->
      function plj ->
	function ml ->
	let nc = ref(0) in
	let nd = ref(0) in
(*	let check_components =
   function nn ->
   function pj ->
   prerr_string("**\n"); flush stderr;
   Logic.prntnpl pj; flush stdout;
   print_string("-----------"^string_of_int(!nd)^"\n");
   let asgs = getmodels pj (ss.kman) nn in
   prerr_string("**\n"); flush stderr;
   asgs
   in
 *)
	let rec join_principals = 
	  function () ->
	    let nk = (ss.init ss.root_node) in
	    if (Node.isNul nk) then []
	    else ( 
	      if (ss.opts.feedback) then (nc:=!nc+1; prerr_string("\b\b\b\b\b\b\b\b\b\b"^string_of_int(!nc)^"/"^string_of_int(ss.initial_nodes)^"   "); flush stderr)
	      else ();
	      let ne = (nk, normalize (Not(pl)) nk, normalize ((plj)) nk ) in ne::(join_principals())
	     )
	in (
(*	let rec first_generation =
   function ls ->
   function js ->
   match ls, js with
   | nn::nns, pj::pjs -> 
   if (ss.opts.feedback) then (nd:=!nd+1; prerr_string("\b\b\b\b\b\b\b\b\b\b"^string_of_int(!nd)^"/"^string_of_int(ss.initial_nodes)^"   "); flush stderr)
   else ();
   let mdls = check_components nn pj in
   let nodes = (ss.join nn) in
   prerr_string("*\n"); flush stderr;
   nodes::(first_generation nns pjs)
   | [], [] ->  []		  
   in
 *)
	let rec first_generation_lazy =
	  function ls ->
	    function js ->
	      match ls, js with
	      | nn::nns, pj::pjs -> 
		  if (ss.opts.feedback) then (nd:=!nd+1; prerr_string("\b\b\b\b\b\b\b\b\b\b"^string_of_int(!nd)^"/"^string_of_int(ss.initial_nodes)^"   "); flush stderr)
		  else ();
		  let nodes = ss.join nn in
		  let thenodes = filter nodes (fast_sat pj (ss.kman)) in
		  let res = map thenodes (function n -> (n,Csolver.init())) in
		  res::(first_generation_lazy nns pjs)
	      | [], [] ->  []		  
	in
	if (ss.opts.feedback) then (
	  prerr_string("\n\nJoining principals...\n\n");
	  flush stderr
	 ) else ();

	let stime = Sys.time() in
        let rd = join_principals() in
	let olist = reorder ml rd in
       	let rb = List.rev (order_list olist) in
	if (ss.opts.feedback) then (
	  prerr_string("\n\nFirst generation...\n\n");
	  flush stderr
	 ) else ();
 	ss.startnodes <- (map rb (function (nn, pp, pj) -> nn));
	ss.startpls <- (map rb (function (nn, pp, pj) -> pp));
	ss.joinpls <- (map rb (function (nn, pp, pj) -> pj));
	ss.current_node <- 0;
	ss.initial_nodes <- len ss.startnodes;
	ss.len_traces <- map ss.startnodes (function nn -> let (ctx,asgx,kx) = Node.getstatus nn in Context.getlentrace ctx);
	ss.num_nodes <- ss.initial_nodes +1;
	ss.transitions <- ss.initial_nodes;
	let ra =
(*	  if (ss.opts.lazyness) then *)
	  first_generation_lazy (ss.startnodes) (ss.joinpls) 
(*	  else first_generation (ss.startnodes) (ss.joinpls) *)
	in
	
        let oVar= List.map (fun x-> varCount x)  (ss.startnodes) in
      
	 let rec verifier nodel valu lf=
           let v = H_var.check_open valu (Utils.debox lf) (0) in	
	   match nodel with 
	   (nn,cs)::nns->
		 let rec verConj n vv fff =
		  match fff with 
		  |Equal(var,m)::fs-> 
		     let checkmsg = H_var.check n var in
		     (match checkmsg with 
		     |Bottom-> 0::verConj n vv fs
	             |PN(name,index)->
			if PN(name,index)=m then 
			       let newv =vv in newv::(verConj n vv fs) 
			else (-1)::(verConj n vv fs)
		     |SVar((va,(name,index)),id,svt)->
			let boolvalue =(Knowledge.knows (Kmanager.getK ss.kman id) m) in 
			  (match boolvalue with
			  true->let newv =vv in newv::verConj n vv fs
			  |false->(-1)::(verConj n vv fs))
			  |_-> 	let  msg = H_var.check n m in 
			     if msg = checkmsg then let newv =vv in newv::(verConj n vv fs) 
			     else (-1)::(verConj n vv fs))
												
			     (*let checkmsg = H_var.check n var in 
			     if (Message.notSVar checkmsg) then
			     (if checkmsg = Bottom then (0::verConj n vv fs ss)
			     else 
			     let  msg = H_var.check n m in 
			     if msg = checkmsg then let newv =vv in 
		             vv::(verConj n newv fs ss) else (-1)::(verConj n vv fs ss))
                             else	
                             (let SVar((va,(name,index)),id,svt) = checkmsg  in 
			     let boolvalue =(Knowledge.knows (Kmanager.getK ss.kman id) m) in 
			      match boolvalue with
			      true->let newv =vv in newv::verConj n vv fs ss
			     |false->(-1)::verConj n vv fs ss)*)
			
		  |Diff(var,m)::fs-> let checkms = H_var.check n var in 
		     (match checkms with 
		     |Bottom-> 0::verConj n vv fs
		     |PN(name,index)-> 
		       if PN(name,index)<>m then 
			    let newv =vv in newv::(verConj n vv fs) 
			else (-1)::(verConj n vv fs)
		     |SVar((va,(name,index)),id,svt)->
		        let boolvalue =(Knowledge.knows (Kmanager.getK ss.kman id) m) in 
			 (match boolvalue with
		         false->let newv =vv in newv::verConj n vv fs
			 |true->let newv = (vv-1) in newv::(verConj n (vv-1) fs))
		         |_->let  msg = H_var.check n m in 
			  if msg <> checkms then let newv =vv in newv::(verConj n vv fs)
			  else (-1)::(verConj n vv fs))
			  (*let checkmsg = H_var.check n var in 
			  if (Message.notSVar checkmsg) then
			  (if checkmsg = Bottom then (0::verConj n vv fs ss)
			  else 
			  let  msg = H_var.check n m in 
			  if msg = checkmsg then 
			  let newv =vv in newv::(verConj n vv fs ss) else (-1)::(verConj n vv fs ss))
			  else	
 			  (let SVar((va,(name,index)),id,svt) = checkmsg  in 
			  let boolvalue =(Knowledge.knows (Kmanager.getK ss.kman id) m) in 
			  match boolvalue with
			  false->let newv= vv in newv::verConj n vv fs ss
			  |true->let newv = (vv-1) in newv::(verConj n (vv-1) fs ss))*)
					
		  |True::fs->1::verConj n vv fs 
		  |False::fs->(-1)::verConj n vv fs
		  |Derive(_)::fs->0::verConj n vv fs 
		  |NDerive(_)::fs->0::verConj n vv fs
		  |[]->vv::[]
	    in                                                                 
	    let rec conj node value f=
	    match f with 
	    ff::ffs-> let newvalue = (verConj node v ff) in
	     let varvalue = List.hd (List.rev newvalue) in
	     (maxm (List.tl (List.rev newvalue)))::(conj node varvalue ffs) 
	    |[]->[]
            				
             in 
             let newmm = (conj nn v lf) in
	     let newm =(H_var.max newmm,(nn,cs))::(verifier nns valu lf) in
	     newm
        |[]->[]					
			
	in			
       	let rec h_2 s vlist g  = 
         match s,g,vlist with 
	 nl::nls,g::gs,vl::vs-> (verifier nl vl g)::(h_2 nls vs gs) 
         |[],[],[]->[] 
				  		 
   	in	
	let rll= h_2 (ra) oVar (ss.startpls) in
        let rl =order rll in
	ss.startjoins <- rl;
	ss.son_traces <- map (ss.startjoins) (function l -> float_of_int(len l));
	ss.jointime <- Sys.time() -. stime
       )
	  

let do_search =
  function ss ->
    let rec searchstack = 
      function ls ->
	match ls with
	| (nn,cf)::nns -> (
	    let nstck = ref(Thestack.init nn) in
	    (
	     while (not (Thestack.isEmpty (!nstck))) do
	       (nstck := (iterstack ss cf (!nstck)))
	     done
	    )); searchstack nns
	| []  -> ()
    in
    let rec startsearch =
      function ls ->
	match ls with
	| jn::jns -> 
	    searchstack jn; ss.current_node <- ss.current_node+1;  feedback ss; startsearch jns
	| [] ->  ss.current_node <- ss.current_node-1;  feedback ss
    in
    let stime = Sys.time() in
    if (ss.opts.feedback) then (
      prerr_string("\n\nSearching...\n\n");
      flush stderr
     ) else ();
    (try (startsearch ss.startjoins) with StopSearch -> ());
    ss.searchtime <- Sys.time() -. stime
	



let prnt_stats =
  function chan ->
    function ss ->
      output_string chan ("\n\n--------------------------------------------\nTotal traces: "^string_of_int(ss.num_traces)^"\n");
      output_string chan ("Total terminal traces: "^string_of_int(ss.num_terminal_traces)^"\n");
      output_string chan ("Total nodes: "^string_of_int(ss.num_nodes)^"\n");
      output_string chan ("Total starting nodes: "^string_of_int(ss.initial_nodes)^"\n");
      output_string chan ("Total pruned starting nodes: "^string_of_int(len (filter ss.startjoins (function l -> (len l) = 0)))^"\n");
      output_string chan ("\nTotal time: "^format_time(ss.searchtime +. ss.jointime)^"\n");
      output_string chan ("Search time: "^format_time(ss.searchtime -. ss.verifytime )^"\n");
      output_string chan ("Join time: "^format_time(ss.jointime)^"\n");
      output_string chan ("Verify time: "^format_time(ss.verifytime)^"\n");
      let nps = float_of_int(ss.num_nodes) /. (ss.searchtime -. ss.verifytime) in
      let nts = float_of_int(ss.num_traces) /. (ss.searchtime -. ss.verifytime) in
      let nvs = float_of_int(ss.num_terminal_traces) /. (ss.verifytime) in
      output_string chan ("\nnode/s: "^string_of_int(int_of_float(nps))^"\n");
      output_string chan ("trace/s: "^string_of_int(int_of_float(nts))^"\n");
      output_string chan ("verification/s: "^string_of_int(int_of_float(nvs))^"\n");
      output_string chan ("branching factor: "^string_of_float(ss.bfm)^", ("^string_of_int(ss.transitions)^"/"^string_of_int(ss.closed_nodes)^")\n");
      output_string chan ("max branching factor: "^string_of_int(ss.maxbf)^"\n");
      output_string chan ("min branching factor: "^string_of_int(ss.minbf)^"\n");
      output_string chan ("initial branching factors: ");
      print_list (ss.son_traces) (function i -> output_string chan (string_of_int(int_of_float(i))^" "));      
      output_string chan ("\ntrace length: ");
      print_list (ss.len_traces) (function i -> output_string chan (string_of_int(i)^" "));
      output_string chan ("\n"); 
      output_string chan ("\nNumber of attacks: "^string_of_int(ss.num_attacks)^"\n--------------------------------------------\n")

let print_stats = prnt_stats stdout
let prerr_stats = prnt_stats stderr
